from abc import ABC, abstractclassmethod
from dataclasses import dataclass
from typing import Any

import tensorflow as tf
from tensorflow.keras.optimizers.schedules import LearningRateSchedule

from ..registry import cfg_serializable


class BaseLRSchedule(ABC):
    def __init__(self, cfg: Any, timekeeping: Any):
        self.cfg = cfg
        self.timekeeping = timekeeping

    @abstractclassmethod
    def __call__(self) -> LearningRateSchedule:
        """Return a tensorflow learning rate schedule"""


@dataclass
class LRConstConfig:
    # Learning rate value
    lr: float


@cfg_serializable
class LRConstFactory(BaseLRSchedule):
    cfg_class = LRConstConfig

    def __init__(self, cfg: LRConstConfig, timekeeping: Any):
        super().__init__(cfg, timekeeping)

    def __call__(self):
        # We simulate a constant lr using a keras `LearningRateSchedule` so we can
        # wrap it using the `WarmupWrapper` below.
        return tf.keras.optimizers.schedules.PiecewiseConstantDecay(
            boundaries=[0], values=[self.cfg.lr, self.cfg.lr]
        )


@dataclass
class LRMultiStepsConfig:
    # Used by `multisteps`. At the epochs given by `lr_boundaries` we change the lr to
    # the values given by `lr_values`. Note that we need
    # `len(lr_values) = len(lr_boundaries) + 1`.
    lr_boundaries: tuple
    lr_values: tuple


@cfg_serializable
class LRMultiStepsFactory(BaseLRSchedule):
    cfg_class = LRMultiStepsConfig

    def __init__(self, cfg: LRMultiStepsConfig, timekeeping: Any):
        super().__init__(cfg, timekeeping)

    def __call__(self):
        # We convert `lr_boundaries` from epochs to steps.
        boundaries = [
            val * self.timekeeping.nb_steps_per_epoch for val in self.cfg.lr_boundaries
        ]
        return tf.keras.optimizers.schedules.PiecewiseConstantDecay(
            boundaries=boundaries, values=self.cfg.lr_values
        )


@dataclass
class LRCosineDecayConfig:
    # Learning rate value
    lr: float
    alpha: float = 0.0


@cfg_serializable
class LRCosineDecayFactory(BaseLRSchedule):
    cfg_class = LRCosineDecayConfig

    def __init__(self, cfg: LRCosineDecayConfig, timekeeping: Any):
        super().__init__(cfg, timekeeping)

    def __call__(self):
        return tf.keras.optimizers.schedules.CosineDecay(
            self.cfg.lr, decay_steps=self.timekeeping.nb_steps, alpha=self.cfg.alpha
        )


@dataclass
class LRExpDecayConfig:
    # Learning rate value
    lr: float
    # Used by `exponential_decay`. We decay the lr every `lr_decay_frequency` epochs
    # by the facto `lr_decay_rate`.
    lr_decay_rate: float
    lr_decay_frequency: int
    staircase: bool = True


@cfg_serializable
class LRExponentialDecayFactory(BaseLRSchedule):
    cfg_class = LRExpDecayConfig

    def __init__(self, cfg: LRExpDecayConfig, timekeeping: Any):
        super().__init__(cfg, timekeeping)

    def __call__(self):
        return tf.keras.optimizers.schedules.ExponentialDecay(
            initial_learning_rate=self.cfg.lr,
            decay_steps=self.cfg.lr_decay_frequency
            * self.timekeeping.nb_steps_per_epoch,
            decay_rate=self.cfg.lr_decay_rate,
            staircase=self.cfg.staircase,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from copy import deepcopy
from typing import Tuple

import numpy as np
import torch
from torch.nn import functional as F
from torchvision.transforms.functional import resize, to_pil_image  # type: ignore


class ResizeLongestSide:
    """
    Resizes images to longest side 'target_length', as well as provides
    methods for resizing coordinates and boxes. Provides methods for
    transforming both numpy array and batched torch tensors.
    """

    def __init__(self, target_length: int) -> None:
        self.target_length = target_length

    def apply_image(self, image: np.ndarray) -> np.ndarray:
        """
        Expects a numpy array with shape HxWxC in uint8 format.
        """
        target_size = self.get_preprocess_shape(
            image.shape[0], image.shape[1], self.target_length
        )
        return np.array(resize(to_pil_image(image), target_size))

    def apply_coords(
        self, coords: np.ndarray, original_size: Tuple[int, ...]
    ) -> np.ndarray:
        """
        Expects a numpy array of length 2 in the final dimension. Requires the
        original image size in (H, W) format.
        """
        old_h, old_w = original_size
        new_h, new_w = self.get_preprocess_shape(
            original_size[0], original_size[1], self.target_length
        )
        coords = deepcopy(coords).astype(float)
        coords[..., 0] = coords[..., 0] * (new_w / old_w)
        coords[..., 1] = coords[..., 1] * (new_h / old_h)
        return coords

    def apply_boxes(
        self, boxes: np.ndarray, original_size: Tuple[int, ...]
    ) -> np.ndarray:
        """
        Expects a numpy array shape Bx4. Requires the original image size
        in (H, W) format.
        """
        boxes = self.apply_coords(boxes.reshape(-1, 2, 2), original_size)
        return boxes.reshape(-1, 4)

    def apply_image_torch(self, image: torch.Tensor) -> torch.Tensor:
        """
        Expects batched images with shape BxCxHxW and float format. This
        transformation may not exactly match apply_image. apply_image is
        the transformation expected by the model.
        """
        # Expects an image in BCHW format. May not exactly match apply_image.
        target_size = self.get_preprocess_shape(
            image.shape[0], image.shape[1], self.target_length
        )
        return F.interpolate(
            image, target_size, mode="bilinear", align_corners=False, antialias=True
        )

    def apply_coords_torch(
        self, coords: torch.Tensor, original_size: Tuple[int, ...]
    ) -> torch.Tensor:
        """
        Expects a torch tensor with length 2 in the last dimension. Requires the
        original image size in (H, W) format.
        """
        old_h, old_w = original_size
        new_h, new_w = self.get_preprocess_shape(
            original_size[0], original_size[1], self.target_length
        )
        coords = deepcopy(coords).to(torch.float)
        coords[..., 0] = coords[..., 0] * (new_w / old_w)
        coords[..., 1] = coords[..., 1] * (new_h / old_h)
        return coords

    def apply_boxes_torch(
        self, boxes: torch.Tensor, original_size: Tuple[int, ...]
    ) -> torch.Tensor:
        """
        Expects a torch tensor with shape Bx4. Requires the original image
        size in (H, W) format.
        """
        boxes = self.apply_coords_torch(boxes.reshape(-1, 2, 2), original_size)
        return boxes.reshape(-1, 4)

    @staticmethod
    def get_preprocess_shape(
        oldh: int, oldw: int, long_side_length: int
    ) -> Tuple[int, int]:
        """
        Compute the output size given input size and target long side length.
        """
        scale = long_side_length * 1.0 / max(oldh, oldw)
        newh, neww = oldh * scale, oldw * scale
        neww = int(neww + 0.5)
        newh = int(newh + 0.5)
        return (newh, neww)



"""
We provide an implementation and pretrained weights for Pooling-based Vision
Transformers (PiT).

Paper: Rethinking Spatial Dimensions of Vision Transformers.
`[arXiv:2103.16302] <https://arxiv.org/abs/2103.16302>`_.

Original pytorch code and weights from
`NAVER AI <https://github.com/naver-ai/pit>`_.

This code has been ported from the
`timm <https://github.com/rwightman/pytorch-image-models>`_ implementation.

The following models are available.

* Models trained on ImageNet-1k

  * ``pit_ti_224``
  * ``pit_xs_224``
  * ``pit_s_224``
  * ``pit_b_224``

* Models trained on ImageNet-1k, using knowledge distillation

  * ``pit_ti_distilled_224``
  * ``pit_xs_distilled_224``
  * ``pit_s_distilled_224``
  * ``pit_b_distilled_224``
"""
# PiT
# Copyright 2021-present NAVER Corp.
# Apache License v2.0
# Modifications for timm by / Copyright 2020 Ross Wightman
# Copyright 2022 Martins Bruveris
from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple, Union

import numpy as np
import tensorflow as tf

from tfimm.architectures.vit import ViTBlock
from tfimm.layers import interpolate_pos_embeddings_grid, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint fn to this
__all__ = ["PoolingVisionTransformer", "PoolingVisionTransformerConfig"]


@dataclass
class PoolingVisionTransformerConfig(ModelConfig):
    """
    Configuration class for ConvNeXt models.

    Parameters:
        name: Name of the model.
        url: URL for pretrained weights.
        nb_classes: Number of classes for classification head.
        in_channels: Number of input image channels.
        input_size: Input image size (height, width)

        patch_size: Patchifying the image is implemented via a convolutional layer with
            kernel size ``patch_size`` and stride given by ``stride``.
        stride: Stride in patch embedding layer.
        embed_dim: Feature dimensions at each stage.
        nb_blocks: Number of blocks at each stage.
        nb_heads: Number of self-attention heads at each stage.
        mlp_ratio: Ratio of mlp hidden dim to embedding dim
        distilled: If ``True``, we add a distillation head in addition to classification
            head.

        drop_rate: Dropout rate.
        attn_drop_rate: Attention dropout rate.
        drop_path_rate: Dropout rate for stochastic depth.

        norm_layer: Normalization layer. See :func:`~norm_layer_factory` for possible
            values.
        act_layer: Activation function. See :func:`~act_layer_factory` for possible
            values.

        interpolate_input: If ``True``, we interpolate position embeddings to given
            input size, so inference can be done for arbitrary input shape. If ``False``
            inference can only be performed at ``input_size``.
        crop_pct: Crop percentage for ImageNet evaluation.
        interpolation: Interpolation method for ImageNet evaluation.
        mean: Defines preprocessing function. If ``x`` is an image with pixel values
            in (0, 1), the preprocessing function is ``(x - mean) / std``.
        std: Defines preprpocessing function.

        first_conv: Name of first convolutional layer. Used by
            :func:`~tfimm.create_model` to adapt the number in input channels when
            loading pretrained weights.
        classifier: Name of classifier layer. Used by :func:`~tfimm.create_model` to
            adapt the classifier when loading pretrained weights.
    """

    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: int = 16
    stride: int = 8
    embed_dim: Tuple = (64, 128, 256)
    nb_blocks: Tuple = (2, 6, 4)
    nb_heads: Tuple = (2, 4, 8)
    mlp_ratio: float = 4.0
    distilled: bool = False
    # Regularization
    drop_rate: float = 0.0
    attn_drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    # Parameters for inference
    interpolate_input: bool = False
    crop_pct: float = 0.9
    interpolation: str = "bicubic"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "patch_embed/conv"
    classifier: Union[str, Tuple[str, str]] = "head"

    @property
    def nb_tokens(self) -> int:
        """Number of special tokens. Equals 2 if distillation is used, otherwise 1."""
        return 2 if self.distilled else 1

    @property
    def grid_size(self) -> Tuple[int, int]:
        """Grid size for patch embeddings."""
        height = (self.input_size[0] - self.patch_size) // self.stride + 1
        width = (self.input_size[1] - self.patch_size) // self.stride + 1
        return height, width

    @property
    def transform_weights(self):
        """
        Dictionary of functions to transform weights when loading them in models with
        different configs.
        """
        return {"pos_embed": PoolingVisionTransformer.transform_pos_embed}


class ConvHeadPooling(tf.keras.layers.Layer):
    def __init__(
        self,
        nb_tokens: int,
        in_channels: int,
        out_channels: int,
        stride: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.nb_tokens = nb_tokens
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.stride = stride

        self.pad = tf.keras.layers.ZeroPadding2D(padding=stride // 2)
        self.conv = tf.keras.layers.Conv2D(
            filters=out_channels,
            kernel_size=stride + 1,
            strides=stride,
            groups=in_channels,
            name="conv",
        )
        self.fc = tf.keras.layers.Dense(units=out_channels, name="fc")

    def call(self, x):
        x, input_size = x
        batch_size, _, nb_channels = tf.unstack(tf.shape(x))

        tokens = x[:, : self.nb_tokens]
        x = x[:, self.nb_tokens :]  # (N, L, C)
        x = tf.reshape(x, (batch_size, *input_size, nb_channels))  # (N, H, W, C)

        x = self.pad(x)
        x = self.conv(x)
        tokens = self.fc(tokens)
        output_size = tf.unstack(tf.shape(x)[1:3])

        x = tf.reshape(x, (batch_size, -1, self.out_channels))
        x = tf.concat([tokens, x], axis=1)

        return x, output_size


@keras_serializable
class PoolingVisionTransformer(tf.keras.Model):
    """
    Class implementing a Pooling-based Vision Transformer (PiT).

    Paper: Rethinking Spatial Dimensions of Vision Transformers.
    `[arXiv:2103.16302] <https://arxiv.org/abs/2103.16302>`_

    Parameters:
        cfg: Configuration class for the model.
        **kwargs: Arguments are passed to ``tf.keras.Model``
    """

    cfg_class = PoolingVisionTransformerConfig

    def __init__(self, cfg: PoolingVisionTransformerConfig, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(**kwargs)
        self.cfg = cfg
        norm_layer = norm_layer_factory(cfg.norm_layer)

        self.patch_embed = tf.keras.layers.Conv2D(
            filters=cfg.embed_dim[0],
            kernel_size=cfg.patch_size,
            strides=cfg.stride,
            name="patch_embed/conv",
        )
        self.pos_embed = None
        self.pos_drop = tf.keras.layers.Dropout(rate=cfg.drop_rate)
        self.cls_token = None

        # Stochastic depth
        dpr = np.linspace(0.0, cfg.drop_path_rate, sum(cfg.nb_blocks))
        dpr = np.split(dpr, np.cumsum(cfg.nb_blocks))

        self.blocks = OrderedDict()
        for j in range(len(cfg.nb_blocks)):
            for k in range(cfg.nb_blocks[j]):
                self.blocks[f"stage_{j}/block_{k}"] = ViTBlock(
                    embed_dim=cfg.embed_dim[j],
                    nb_heads=cfg.nb_heads[j],
                    mlp_ratio=cfg.mlp_ratio,
                    qkv_bias=True,
                    drop_rate=cfg.drop_rate,
                    attn_drop_rate=cfg.attn_drop_rate,
                    drop_path_rate=dpr[j][k],
                    norm_layer=cfg.norm_layer,
                    act_layer=cfg.act_layer,
                    name=f"transformers/{j}/blocks/{k}",
                )
            if j < len(cfg.nb_blocks) - 1:
                self.blocks[f"stage_{j}/pool"] = ConvHeadPooling(
                    nb_tokens=cfg.nb_tokens,
                    in_channels=cfg.embed_dim[j],
                    out_channels=cfg.embed_dim[j + 1],
                    stride=2,
                    name=f"transformers/{j}/pool",
                )

        self.norm = norm_layer(name="norm")

        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )
        if cfg.distilled:
            self.head_dist = (
                tf.keras.layers.Dense(units=cfg.nb_classes, name="head_dist")
                if cfg.nb_classes > 0
                else tf.keras.layers.Activation("linear")  # Identity layer
            )
        else:
            self.head_dist = None

    def build(self, input_shape):
        height = (input_shape[1] - self.cfg.patch_size) // self.cfg.stride + 1
        width = (input_shape[2] - self.cfg.patch_size) // self.cfg.stride + 1
        self.pos_embed = self.add_weight(
            # We keep PT style NCHW order to make weight translation easier
            shape=(1, self.cfg.embed_dim[0], height, width),
            initializer=tf.keras.initializers.TruncatedNormal(mean=0.0, stddev=0.02),
            trainable=True,
            name="pos_embed",
        )
        self.cls_token = self.add_weight(
            shape=(1, self.cfg.nb_tokens, self.cfg.embed_dim[0]),
            initializer=tf.keras.initializers.TruncatedNormal(mean=0.0, stddev=0.02),
            trainable=True,
            name="cls_token",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        """Returns a tensor of the correct shape for inference."""
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        """
        Names of features, returned when calling ``call`` with ``return_features=True``.
        """
        _, features = self(self.dummy_inputs, return_features=True)
        return list(features.keys())

    def transform_pos_embed(
        self, src_weights, target_cfg: PoolingVisionTransformerConfig
    ):
        """
        Transforms the position embedding weights in accordance with `target_cfg` and
        returns them.
        """
        pos_embed = interpolate_pos_embeddings_grid(
            pos_embed=tf.transpose(self.pos_embed, [0, 2, 3, 1]),
            tgt_grid_size=target_cfg.grid_size,
        )
        pos_embed = tf.transpose(pos_embed, [0, 3, 1, 2])
        return pos_embed

    def forward_features(
        self, x, training: bool = False, return_features: bool = False
    ):
        """
        Forward pass through model, excluding the classifier layer. This function is
        useful if the model is used as input for downstream tasks such as object
        detection.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.patch_embed(x)
        pos_embed = tf.transpose(self.pos_embed, [0, 2, 3, 1])
        if not self.cfg.interpolate_input:
            x = x + pos_embed
        else:
            grid_size = tf.unstack(tf.shape(x)[1:3])
            pos_embed = interpolate_pos_embeddings_grid(
                pos_embed,
                tgt_grid_size=grid_size,
            )
            x = x + pos_embed
        x = self.pos_drop(x, training=training)

        batch_size, height, width, nb_channels = tf.unstack(tf.shape(x))
        input_size = (height, width)
        cls_token = tf.repeat(self.cls_token, repeats=batch_size, axis=0)
        x = tf.reshape(x, (batch_size, -1, nb_channels))
        x = tf.concat([cls_token, x], axis=1)
        features["patch_embedding"] = x

        for key, block in self.blocks.items():
            if key.endswith("pool"):
                x, input_size = block((x, input_size), training=training)
            else:
                x = block(x, training=training)
            features[key] = x
        features["features_all"] = x

        x = x[:, : self.cfg.nb_tokens]
        x = self.norm(x, training=training)
        x = x if self.cfg.distilled else x[:, 0]
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training: bool = False, return_features: bool = False):
        """
        Forward pass through the full model.

        Arguments:
             x: Input to model
             training: Training or inference phase?
             return_features: If ``True``, we return not only the model output, but a
                dictionary with intermediate features.

        Returns:
            If ``return_features=True``, we return a tuple ``(y, features)``, where
            ``y`` is the model output and ``features`` is a dictionary with
            intermediate features.

            If ``return_features=False``, we return only ``y``.
        """
        features = OrderedDict()
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x

        if not self.cfg.distilled:
            x = self.head(x, training=training)
        else:
            y = self.head(x[:, 0], training=training)
            y_dist = self.head_dist(x[:, 1], training=training)
            x = tf.stack((y, y_dist), axis=1)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def pit_ti_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_ti_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(64, 128, 256),
        nb_blocks=(2, 6, 4),
        nb_heads=(2, 4, 8),
        mlp_ratio=4.0,
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_xs_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_xs_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(96, 192, 384),
        nb_blocks=(2, 6, 4),
        nb_heads=(2, 4, 8),
        mlp_ratio=4.0,
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_s_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_s_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(144, 288, 576),
        nb_blocks=(2, 6, 4),
        nb_heads=(3, 6, 12),
        mlp_ratio=4.0,
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_b_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_b_224",
        url="[timm]",
        patch_size=14,
        stride=7,
        embed_dim=(256, 512, 1024),
        nb_blocks=(3, 6, 4),
        nb_heads=(4, 8, 16),
        mlp_ratio=4.0,
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_ti_distilled_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_ti_distilled_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(64, 128, 256),
        nb_blocks=(2, 6, 4),
        nb_heads=(2, 4, 8),
        mlp_ratio=4.0,
        distilled=True,
        classifier=("head", "head_dist"),
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_xs_distilled_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_xs_distilled_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(96, 192, 384),
        nb_blocks=(2, 6, 4),
        nb_heads=(2, 4, 8),
        mlp_ratio=4.0,
        distilled=True,
        classifier=("head", "head_dist"),
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_s_distilled_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_s_distilled_224",
        url="[timm]",
        patch_size=16,
        stride=8,
        embed_dim=(144, 288, 576),
        nb_blocks=(2, 6, 4),
        nb_heads=(3, 6, 12),
        mlp_ratio=4.0,
        distilled=True,
        classifier=("head", "head_dist"),
    )
    return PoolingVisionTransformer, cfg


@register_model
def pit_b_distilled_224():
    cfg = PoolingVisionTransformerConfig(
        name="pit_b_distilled_224",
        url="[timm]",
        patch_size=14,
        stride=7,
        embed_dim=(256, 512, 1024),
        nb_blocks=(3, 6, 4),
        nb_heads=(4, 8, 16),
        mlp_ratio=4.0,
        distilled=True,
        classifier=("head", "head_dist"),
    )
    return PoolingVisionTransformer, cfg



import tensorflow as tf


@tf.keras.utils.register_keras_serializable()
class EmbeddingModel(tf.keras.Model):
    """
    This model can be used for embedding learning tasks. It adds to any TFIMM backbone
    a fully connected layer without bias followed by batch norm. Possible applications
    include face recognition.
    """

    def __init__(self, backbone: tf.keras.Model, embed_dim: int, **kwargs):
        super().__init__(**kwargs)
        self.backbone = backbone
        self.embed_dim = embed_dim

        self.fc = tf.keras.layers.Dense(embed_dim, name="emb/fc")
        self.bn = tf.keras.layers.BatchNormalization(
            axis=-1, scale=False, name="emb/bn"
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return self.backbone.dummy_inputs

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.backbone.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.fc(x)
        x = self.bn(x, training=training)
        features["embeddings"] = x
        return (x, features) if return_features else x

    def get_config(self):
        return {
            "backbone": tf.keras.utils.serialize_keras_object(self.backbone),
            "embed_dim": self.embed_dim,
        }

    @classmethod
    def from_config(cls, config):
        return EmbeddingModel(
            backbone=tf.keras.models.model_from_config(config["backbone"]),
            embed_dim=config["embed_dim"],
        )



import math

import tensorflow as tf


class FanoutInitializer(tf.keras.initializers.Initializer):
    """
    Fanout initializer as used by EfficientNet models.

    Args:
        nb_groups: Number of groups for groupwise convolutions. Fanout is performed
            for each group individually.
        depthwise: If True, we set the number of groups equal to the number of input
            channels. This overrides any value for nb_groups.
    """

    def __init__(self, nb_groups: int = 1, depthwise: bool = False):
        self.nb_groups = nb_groups
        self.depthwise = depthwise

    def __call__(self, shape, dtype=None, **kwargs):
        out_channels = shape[-1]
        fan_out = shape[0] * shape[1] * out_channels  # kernel_size * out_channels
        # Set nb_groups to in_channels for depthwise convolutions
        nb_groups = shape[-2] if self.depthwise else self.nb_groups
        fan_out /= nb_groups
        return tf.random.normal(
            shape, mean=0.0, stddev=math.sqrt(2.0 / fan_out), dtype=dtype
        )

    def get_config(self):
        return {"nb_groups": self.nb_groups, "depthwise": self.depthwise}



import logging
import re
from copy import deepcopy
from typing import Callable, Optional

import tensorflow as tf
from tensorflow.python.keras import backend as K

from tfimm.models.registry import is_model, model_class, model_config
from tfimm.utils import cached_model_path, load_pth_url_weights, load_timm_weights


def create_model(
    model_name: str,
    pretrained: bool = False,
    model_path: str = "",
    *,
    in_channels: Optional[int] = None,
    nb_classes: Optional[int] = None,
    **kwargs,
) -> tf.keras.Model:
    """
    Creates a model.

    Args:
        model_name: Name of model to instantiate
        pretrained: If ``True``, load pretrained weights as specified by the ``url``
            field in config. We will check the cache first and download weights only
            if they cannot be found in the cache.

            If ``url`` is ``[timm]``, the weights will be downloaded from ``timm`` and
            converted to TensorFlow. Requires ``timm`` and ``torch`` to be installed.
            If ``url`` starts with ``[pytorch]``, the weights are in PyTorch format
            and ``torch`` needs to be installed to convert them.
        model_path: Path of model weights to load after model is initialized. This takes
            over ``pretrained``.
        in_channels: Number of input channels for model. If ``None``, use default
            provided by model.
        nb_classes: Number of classes for classifier. If set to 0, no classifier is
            used and last layer is pooling layer. If ``None``, use default provided by
            model.
        **kwargs: Other kwargs are model specific.

    Returns:
        The created model.
    """
    if not is_model(model_name):
        raise RuntimeError(f"Unknown model {model_name}.")

    cls = model_class(model_name)
    cfg = model_config(model_name)

    if model_path:
        loaded_model = tf.keras.models.load_model(model_path, compile=False)
    elif pretrained:
        # First try loading model from cache
        model_path = cached_model_path(model_name)
        if model_path:
            loaded_model = tf.keras.models.load_model(model_path)
        elif cfg.url.startswith("[timm]"):
            loaded_model = cls(cfg)
            loaded_model(loaded_model.dummy_inputs)
            # Url can be "[timm]timm_model_name" or "[timm]" in which case we default
            # to model_name.
            timm_model_name = cfg.url[len("[timm]") :] or model_name
            load_timm_weights(loaded_model, timm_model_name)
        elif cfg.url.startswith("[pytorch]"):
            url = cfg.url[len("[pytorch]") :]
            loaded_model = cls(cfg)
            loaded_model(loaded_model.dummy_inputs)
            load_pth_url_weights(loaded_model, url)
        else:
            raise NotImplementedError(
                "Model not found in cache. Download of weights only implemented for "
                "PyTorch models."
            )
    else:
        loaded_model = None

    # Update config with kwargs
    cfg = deepcopy(cfg)
    for key, value in kwargs.items():
        if hasattr(cfg, key):
            setattr(cfg, key, value)
        else:
            logging.warning(
                f"Config for {model_name} does not have field `{key}`. Ignoring field."
            )
    if in_channels is not None:
        setattr(cfg, "in_channels", in_channels)
    if nb_classes is not None:
        setattr(cfg, "nb_classes", nb_classes)

    # `keras.Model` kwargs need separate treatment. For now we support only `name`.
    model_kwargs = {}
    if "name" in kwargs:
        model_kwargs["name"] = kwargs["name"]

    # If we have loaded a model and the model has the correct config, then we are done.
    if loaded_model is not None and loaded_model.cfg == cfg:
        return loaded_model

    # Otherwise, we build a new model and transfer the weights to it. This is because
    # some parameter changes (in_channels and nb_classes) require changing the shape of
    # some weights or dropping of others. And there might be non-trivial interactions
    # between various parameters, e.g., global_pool can be None only if nb_classes is 0.
    model = cls(cfg, **model_kwargs)
    model(model.dummy_inputs)  # Call model to build layers

    # Now we need to transfer weights from loaded_model to model
    if loaded_model is not None:
        transfer_weights(loaded_model, model)

    return model


def create_preprocessing(
    model_name: str,
    *,
    in_channels: Optional[float] = None,
    dtype: Optional[str] = None,
) -> Callable:
    """
    Creates a function to preprocess images for a particular model.

    The input to the preprocessing function is assumed to be values in range [0, 255].

    Args:
        model_name: Model for which to create preprocessing function.
        in_channels: Number of input channels to model
        dtype: Output dtype.

    Returns:
        Callable that operates on single images as well as batches.
    """
    if not is_model(model_name):
        raise ValueError(f"Unknown model: {model_name}.")

    cfg = model_config(model_name)
    dtype = dtype or tf.keras.backend.floatx()

    def _adapt_vector(v, n):
        """Adapts vector v to length n by repeating as necessary."""
        v = tf.convert_to_tensor(v, dtype=dtype)
        m = tf.shape(v)[0]
        nb_repeats = n // m + 1
        v = tf.tile(v, [nb_repeats])
        v = v[:n]
        return v

    in_channels = in_channels or cfg.in_channels
    mean = _adapt_vector(cfg.mean, in_channels)
    std = _adapt_vector(cfg.std, in_channels)

    def _preprocess(img: tf.Tensor) -> tf.Tensor:
        img = tf.cast(img, dtype=dtype) / 255.0
        img = (img - mean) / std
        return img

    return _preprocess


def transfer_weights(src_model: tf.keras.Model, dst_model: tf.keras.Model):
    """
    Transfers weights from ``src_model`` to ``dst_model``, with special treatment of
    first convolution and classification layers. The ``dst_model`` is modified in place.

    Args:
        src_model: Source model.
        dst_model: Destination model, modified in place
    """
    dst_first_conv = dst_model.cfg.first_conv

    if hasattr(src_model.cfg, "nb_classes") and hasattr(dst_model.cfg, "nb_classes"):
        keep_classifier = src_model.cfg.nb_classes == dst_model.cfg.nb_classes
    else:
        # It doesn't really matter what we set this to, since if a model doesn't have
        # a nb_classes parameter, it most likely won't have a classifier parameter
        # set either
        keep_classifier = True
    dst_classifier = getattr(dst_model.cfg, "classifier", [])
    if isinstance(dst_classifier, str):  # Some models have multiple classifier heads
        dst_classifier = [dst_classifier]

    # For some weights we may have to apply custom transforms to adapt them to the new
    # model config
    transform_weights = getattr(src_model.cfg, "transform_weights", dict())

    # Some weights should not be transferred as they are created during building
    weights_to_ignore = getattr(dst_model, "keys_to_ignore_on_load_missing", [])

    src_weights = {_get_weight_name(w.name): w.numpy() for w in src_model.weights}
    weight_value_tuples = []
    for dst_weight in dst_model.weights:
        # We store human-friendly names in the model config.
        layer_name = _get_layer_name(dst_weight.name)
        w_name = _get_weight_name(dst_weight.name)

        if any(re.search(pat, w_name) is not None for pat in weights_to_ignore):
            # Weights that don't need to be transferred
            pass

        elif layer_name in dst_classifier:
            if keep_classifier:
                # We only keep the classifier if the number of classes is the same
                # Otherwise the classifier is not copied over, i.e., we keep the
                # initialization of dst_model.
                weight_value_tuples.append((dst_weight, src_weights[w_name]))

        elif layer_name == dst_first_conv:
            src_weight = _transform_first_conv(
                src_weights[w_name], dst_model.cfg.in_channels
            )
            weight_value_tuples.append((dst_weight, src_weight))

        elif w_name in transform_weights:
            # We check if we need to apply a transform. In that case we call the
            # transform function passing the source model, source weight and target
            # config.
            src_weight = transform_weights[w_name](
                src_model, src_weights[w_name], dst_model.cfg
            )
            weight_value_tuples.append((dst_weight, src_weight))

        else:
            # All other weights are simply copied over
            weight_value_tuples.append((dst_weight, src_weights[w_name]))

    # This modifies weights in place
    K.batch_set_value(weight_value_tuples)


def _get_layer_name(name):
    """
    Extracts the name of the layer, which is compared against the config values for
    `first_conv` and `classifier`.

    The input is, e.g., name="res_net/remove/fc/kernel:0". We want to extract "fc".
    """
    name = name.replace(":0", "")  # Remove device IDs
    name = name.replace("/remove/", "/")  # Auxiliary intermediate levels
    # The model name prefix is made unique by TF, i.e., two ResNets will have variables
    # 'res_net/var:0' and 'res_net_1/var:0'. Here we remove the prefix.
    name = name.split("/", 1)[-1]  # Remove prefix, e.g., "res_net"
    name = name.rsplit("/", 1)[0]  # Remove last part, e.g., "kernel"
    return name


def _get_weight_name(name):
    """
    Extracts the name of the weight, which is compared against the config values for
    `transfer_weights`. The difference with `_get_weights_name` is that we preserve
    the last part.

    The input is, e.g., name="res_net/fc/kernel:0". We want to extract "fc/kernel".
    """
    name = name.replace(":0", "")  # Remove device IDs
    name = name.split("/", 1)[-1]  # Remove prefix, e.g., "res_net"
    return name


def _transform_first_conv(weight, in_channels):
    """
    Adapts `weight` to have `in_channels` either by truncating or by repeating
    across channel dimensions.
    """
    if tf.rank(weight) != 4:
        # We don't need to adapt biases, because they don't depend on input channels
        return weight

    # Convolutional kernels have shape (H, W, in_channels, out_channels)
    src_channels = tf.shape(weight)[2]
    if in_channels == src_channels:
        # Nothing to adapt here...
        pass
    elif in_channels == 1:
        # For single-channel input, we sum the existing channels. We don't average in
        # order to preserve weight statistics
        weight = tf.reduce_sum(weight, axis=2, keepdims=True)
    else:
        nb_repeats = in_channels // src_channels + 1
        weight = tf.tile(weight, [1, 1, nb_repeats, 1])
        weight = weight[:, :, :in_channels, :]
        weight *= tf.cast(src_channels / in_channels, tf.float32)
    return weight



"""
Common layers shared between transformer architectures.

Copyright 2021 Martins Bruveris
"""
from typing import Optional, Tuple

import tensorflow as tf

from tfimm.layers.factory import act_layer_factory, norm_layer_factory


def interpolate_pos_embeddings(
    pos_embed: tf.Tensor,
    src_grid_size: Tuple[int, int],
    tgt_grid_size: Tuple[int, int],
    nb_tokens: int = 0,
) -> tf.Tensor:
    """
    This method allows to interpolate the pre-trained position encodings, to be
    able to use the model on higher resolution images.

    Args:
        pos_embed: Positional embeddings to interpolate, shape (1, N, D)
        src_grid_size: Grid size of given embeddings.
        tgt_grid_size: Input size to which position embeddings should be adapted
        nb_tokens: How many token should be ignored for interpolation (e.g., class or
            distillation tokens)

    Returns:
        Position embeddings (including class tokens) appropriate to input_size
    """
    if src_grid_size == tgt_grid_size:
        return pos_embed  # No interpolation needed

    src_pos_embed = pos_embed[:, nb_tokens:]
    src_pos_embed = tf.reshape(src_pos_embed, shape=(1, *src_grid_size, -1))
    tgt_pos_embed = tf.image.resize(
        images=src_pos_embed,
        size=tgt_grid_size,
        method="bicubic",
    )
    tgt_pos_embed = tf.cast(tgt_pos_embed, dtype=src_pos_embed.dtype)
    nb_pos_tokens = tgt_grid_size[0] * tgt_grid_size[1]
    tgt_pos_embed = tf.reshape(tgt_pos_embed, shape=(1, nb_pos_tokens, -1))
    tgt_pos_embed = tf.concat((pos_embed[:, :nb_tokens], tgt_pos_embed), axis=1)
    return tgt_pos_embed


def interpolate_pos_embeddings_grid(
    pos_embed: tf.Tensor,
    tgt_grid_size: Tuple[int, int],
) -> tf.Tensor:
    """
    This method allows to interpolate the pre-trained position encodings, to be
    able to use the model on higher resolution images. We use this function if position
    embeddings are already given as a grid.

    Args:
        pos_embed: Positional embeddings to interpolate, shape (B, H, W, D)
        tgt_grid_size: Input size to which position embeddings should be adapted

    Returns:
        Position embeddings (including class tokens) appropriate to input_size
    """
    height, width = tf.unstack(tf.shape(pos_embed)[1:3])
    if (height, width) == tgt_grid_size:
        return pos_embed  # No interpolation needed

    tgt_pos_embed = tf.image.resize(
        images=pos_embed,
        size=tgt_grid_size,
        method="bicubic",
    )
    tgt_pos_embed = tf.cast(tgt_pos_embed, dtype=pos_embed.dtype)
    return tgt_pos_embed


class PatchEmbeddings(tf.keras.layers.Layer):
    """
    Image to Patch Embedding.

    Supports overlapping patches when stride is specified. Used, e.g., in Pyramid
    Vision Transformer V2 or PoolFormer.

    Parameters:
        patch_size: Patchifying the image is implemented via a convolutional layer with
            kernel size equal to ``patch_size``.
        embed_dim: Number of embedding dimensions.
        stride: If ``None``, we use non-overlapping patches, i.e.,
            ``stride=patch_size``. Other values are used e.g., in PVT v2 or PoolFormer.
        padding: Padding is only applied when overlapping patches are used. In that
            case we default to ``padding = patch_size // 2``, but other values can be
            specified via this parameter. For non-overlapping patches, padding is
            always 0.
        flatten: If ``True``, we reshape the patchified image from ``(H', W', D)`` into
            ``(H'*W', D)``.
        norm_layer: Normalization layer to be applied after the patch projection.
        kernel_initializer: Initializer for kernel weights
        bias_initializer: Initializer for bias weights
        **kwargs: Other arguments are passed to the parent class.
    """

    def __init__(
        self,
        patch_size: int,
        embed_dim: int,
        stride: Optional[int] = None,
        padding: Optional[int] = None,
        flatten: bool = True,
        norm_layer: str = "",
        kernel_initializer: str = "glorot_uniform",
        bias_initializer: str = "zeros",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        # If stride=None we default to non-overlapping patches
        self.stride = stride or patch_size
        # Default padding for PVT transformer
        self.padding = padding or patch_size // 2
        self.flatten = flatten
        self.norm_layer = norm_layer_factory(norm_layer)

        # We only apply padding, if we use overlapping patches. For non-overlapping
        # patches we assume image size is divisible by patch size.
        self.pad = tf.keras.layers.ZeroPadding2D(
            padding=self.padding if self.stride != self.patch_size else 0
        )
        self.projection = tf.keras.layers.Conv2D(
            filters=self.embed_dim,
            kernel_size=self.patch_size,
            strides=self.stride,
            use_bias=True,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="proj",
        )
        self.norm = self.norm_layer(name="norm")

    def call(self, x, training=False, return_shape=False):
        """
        Forward pass through the layer.

        An image of shape ``(H, W, C)`` will be mapped to patches of shape
        ``(H', W', D)``, where ``D`` is ``embed_dim``. The output is then optionally
        flattened to ``(H'*W', D)``.

        Arguments:
             x: Input to layer
             training: Training or inference phase?
             return_shape: If ``True``, we additionally return the spatial shape of
                the image as well as the tokens.

        Returns:
            We return the flattened tokens and if ``return_shape=True``, additionally
            the shape ``(H', W')``.

            This information is used by models that use convolutional layers in addition
            to attention layers and convolutional layers need to know the original shape
            of the token list.
        """
        x = self.pad(x)
        x = self.projection(x)

        batch_size, height, width = tf.unstack(tf.shape(x)[:3])
        if self.flatten:
            # Change the 2D spatial dimensions to a single temporal dimension.
            x = tf.reshape(tensor=x, shape=(batch_size, height * width, -1))

        x = self.norm(x, training=training)
        return (x, (height, width)) if return_shape else x


class MLP(tf.keras.layers.Layer):
    """MLP as used in Vision Transformer, MLP-Mixer and related networks"""

    def __init__(
        self,
        hidden_dim: int,
        embed_dim: int,
        drop_rate: float,
        act_layer: str,
        kernel_initializer: str = "glorot_uniform",
        bias_initializer: str = "zeros",
        **kwargs,
    ):
        super().__init__(**kwargs)
        act_layer = act_layer_factory(act_layer)

        self.fc1 = tf.keras.layers.Dense(
            units=hidden_dim,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="fc1",
        )
        self.act = act_layer()
        self.drop1 = tf.keras.layers.Dropout(rate=drop_rate)
        self.fc2 = tf.keras.layers.Dense(
            units=embed_dim,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="fc2",
        )
        self.drop2 = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x, training=training)
        x = self.fc2(x)
        x = self.drop2(x, training=training)
        return x


class ConvMLP(tf.keras.layers.Layer):
    """
    ConvMLP is the same block as MLP, but it uses 1x1 convolutions instead of fully
    connected layers.

    Used in ``ConvNeXt`` models.
    """

    def __init__(
        self,
        hidden_dim: int,
        embed_dim: int,
        drop_rate: float,
        act_layer: str,
        kernel_initializer: str = "glorot_uniform",
        bias_initializer: str = "zeros",
        **kwargs,
    ):
        super().__init__(**kwargs)
        act_layer = act_layer_factory(act_layer)

        self.fc1 = tf.keras.layers.Conv2D(
            filters=hidden_dim,
            kernel_size=1,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="fc1",
        )
        self.act = act_layer()
        self.drop1 = tf.keras.layers.Dropout(rate=drop_rate)
        self.fc2 = tf.keras.layers.Conv2D(
            filters=embed_dim,
            kernel_size=1,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="fc2",
        )
        self.drop2 = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x, training=training)
        x = self.fc2(x)
        x = self.drop2(x, training=training)
        return x


class GatedBiasInitializer(tf.keras.initializers.Initializer):
    """
    Splits tensor in half along last axis (channels). Initializes second half with
    ones.

    Used for bias term in Gated Linear Units.
    """

    def __init__(self, initializer="zeros"):
        self.initializer = tf.keras.initializers.get(initializer)

    def __call__(self, shape, dtype=None, **kwargs):
        if dtype is None:
            dtype = tf.keras.backend.floatx()

        assert shape[-1] % 2 == 0
        split_shape = shape[:-1] + [
            shape[-1] // 2,
        ]
        x1 = self.initializer(split_shape, dtype=dtype)
        x2 = tf.ones(split_shape, dtype=dtype)
        x = tf.concat([x1, x2], axis=-1)
        return x


class GatedKernelInitializer(tf.keras.initializers.Initializer):
    """
    Splits tensor in half along last axis (channels). Initializes second half with
    normal distribution with stddev=1e-6.

    Used for kernel term in Gated Linear Units.
    """

    def __init__(self, initializer="glorot_uniform"):
        self.normal = tf.keras.initializers.RandomNormal(stddev=1e-6)
        self.initializer = tf.keras.initializers.get(initializer)

    def __call__(self, shape, dtype=None, **kwargs):
        if dtype is None:
            dtype = tf.keras.backend.floatx()

        assert shape[-1] % 2 == 0
        split_shape = shape[:-1] + [
            shape[-1] // 2,
        ]
        x1 = self.initializer(split_shape, dtype=dtype)
        x2 = self.normal(split_shape, dtype=dtype)
        x = tf.concat([x1, x2], axis=-1)
        return x


class GluMLP(tf.keras.layers.Layer):
    """
    MLP w/ GLU style gating
    See: https://arxiv.org/abs/1612.08083, https://arxiv.org/abs/2002.05202
    """

    def __init__(
        self,
        hidden_dim: int,
        embed_dim: int,
        drop_rate: float,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        act_layer = act_layer_factory(act_layer)
        assert hidden_dim % 2 == 0

        self.fc1 = tf.keras.layers.Dense(
            units=hidden_dim,
            bias_initializer=GatedBiasInitializer(),
            kernel_initializer=GatedKernelInitializer(),
            name="fc1",
        )
        self.act = act_layer()
        self.drop1 = tf.keras.layers.Dropout(rate=drop_rate)
        self.fc2 = tf.keras.layers.Dense(units=embed_dim, name="fc2")
        self.drop2 = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False):
        x = self.fc1(x)
        x, gates = tf.split(x, num_or_size_splits=2, axis=-1)
        x = x * self.act(gates)
        x = self.drop1(x, training=training)
        x = self.fc2(x)
        x = self.drop2(x, training=training)
        return x


class SpatialGatingUnit(tf.keras.layers.Layer):
    """
    Spatial Gating Unit

    Based on: Pay Attention to MLPs - https://arxiv.org/abs/2105.08050
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        norm_layer = norm_layer_factory("layer_norm")
        self.norm = norm_layer(name="norm")

    def build(self, input_shape):
        seq_len = input_shape[-2]
        self.proj = tf.keras.layers.Dense(
            units=seq_len,
            kernel_initializer=tf.keras.initializers.RandomNormal(stddev=1e-6),
            bias_initializer=tf.keras.initializers.Ones(),
            name="proj",
        )

    def call(self, x, training=False):
        u, v = tf.split(x, num_or_size_splits=2, axis=-1)
        v = self.norm(v, training=training)
        v = tf.transpose(v, perm=(0, 2, 1))
        v = self.proj(v)
        v = tf.transpose(v, perm=(0, 2, 1))
        x = u * v
        return x


class GatedMLP(tf.keras.layers.Layer):
    """MLP as used in gMLP"""

    def __init__(
        self,
        hidden_dim: int,
        embed_dim: int,
        drop_rate: float,
        act_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        act_layer = act_layer_factory(act_layer)

        self.fc1 = tf.keras.layers.Dense(units=hidden_dim, name="fc1")
        self.act = act_layer()
        self.drop1 = tf.keras.layers.Dropout(rate=drop_rate)
        self.gate = SpatialGatingUnit(name="gate")
        self.fc2 = tf.keras.layers.Dense(units=embed_dim, name="fc2")
        self.drop2 = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x, training=training)
        x = self.gate(x, training=training)
        x = self.fc2(x)
        x = self.drop2(x, training=training)
        return x



"""
Class-Attention in Image Transformers (CaiT)

Paper: Going deeper with Image Transformers
Link: https://arxiv.org/abs/2103.17239

Based on timm/models/cait.py by Ross Wightman.
Original code and weights from https://github.com/facebookresearch/deit

Copyright 2021 Martins Bruveris
Copyright 2021 Ross Wightman
Copyright 2015-present, Facebook, Inc.
"""
from dataclasses import dataclass
from typing import List, Tuple

import tensorflow as tf

from tfimm.layers import (
    MLP,
    DropPath,
    PatchEmbeddings,
    interpolate_pos_embeddings,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# Model registry will add each entrypoint function to this
__all__ = ["CaiT", "CaiTConfig"]


@dataclass
class CaiTConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: int = 16
    embed_dim: int = 768
    nb_blocks: int = 12
    nb_heads: int = 12
    mlp_ratio: float = 4.0
    qkv_bias: bool = True
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    attn_drop_rate: float = 0.0
    # Other parameters
    norm_layer: str = "layer_norm_eps_1e-6"
    act_layer: str = "gelu"
    init_scale: float = 1e-4
    # Parameters for inference
    interpolate_input: bool = False
    crop_pct: float = 1.0
    interpolation: str = "bicubic"
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "patch_embed/proj"
    classifier: str = "head"

    """
    Args:
        nb_classes: Number of classes for classification head
        in_channels: Number of input channels
        input_size: Input image size
        patch_size: Patch size; Image size must be multiple of patch size
        embed_dim: Embedding dimension
        nb_blocks: Depth of transformer (number of encoder blocks)
        nb_heads: Number of self-attention heads
        mlp_ratio: Ratio of mlp hidden dim to embedding dim
        qkv_bias: Enable bias for qkv if True
        drop_rate: Dropout rate
        attn_drop_rate: Attention dropout rate
        drop_path_rate: Dropout rate for stochastic depth
        norm_layer: Normalization layer
        act_layer: Activation function
        init_scale: Inital value for layer scale weights
        """

    @property
    def grid_size(self) -> Tuple[int, int]:
        return (
            self.input_size[0] // self.patch_size,
            self.input_size[1] // self.patch_size,
        )

    @property
    def nb_patches(self) -> int:
        return self.grid_size[0] * self.grid_size[1]

    @property
    def transform_weights(self):
        return {"pos_embed": CaiT.transform_pos_embed}


class ClassAttention(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        qkv_bias: bool,
        attn_drop_rate: float,
        proj_drop_rate: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.nb_heads = nb_heads
        self.scale = (embed_dim // nb_heads) ** -0.5

        self.q = tf.keras.layers.Dense(units=embed_dim, use_bias=qkv_bias, name="q")
        self.k = tf.keras.layers.Dense(units=embed_dim, use_bias=qkv_bias, name="k")
        self.v = tf.keras.layers.Dense(units=embed_dim, use_bias=qkv_bias, name="v")
        self.attn_drop = tf.keras.layers.Dropout(rate=attn_drop_rate)
        self.proj = tf.keras.layers.Dense(units=embed_dim, name="proj")
        self.proj_drop = tf.keras.layers.Dropout(rate=proj_drop_rate)

    def call(self, x, training=False):
        # B (batch size), N (sequence length), D (embedding dimension),
        # H (number of heads)
        batch_size, seq_length = tf.unstack(tf.shape(x)[:2])
        q = self.q(x[:, 0])  # (B, D)
        q = tf.expand_dims(q, axis=1)  # (B, 1, D)
        q = tf.reshape(q, (batch_size, 1, self.nb_heads, -1))  # (B, 1, H, D/H)
        q = tf.transpose(q, (0, 2, 1, 3))  # (B, H, 1, B/H)
        q = q * self.scale  # (B, H, 1, B/H)

        k = self.k(x)  # (B, N, D)
        k = tf.reshape(k, (batch_size, seq_length, self.nb_heads, -1))  # (B, N, H, D/H)
        k = tf.transpose(k, (0, 2, 1, 3))  # (B, H, N, D/H)

        v = self.v(x)  # (B, N, D)
        v = tf.reshape(v, (batch_size, seq_length, self.nb_heads, -1))  # (B, N, H, D/H)
        v = tf.transpose(v, (0, 2, 1, 3))  # (B, H, N, D/H)

        attn = tf.linalg.matmul(q, k, transpose_b=True)  # (B, H, 1, N)
        attn = tf.nn.softmax(attn, axis=-1)  # (B, H, 1, N)
        attn = self.attn_drop(attn, training=training)  # (B, H, 1, N)

        x = tf.linalg.matmul(attn, v)  # (B, H, 1, D/H)
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, 1, H, D/H)
        x = tf.reshape(x, (batch_size, 1, -1))  # (B, 1, D)

        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return x


class LayerScaleBlockClassAttention(tf.keras.layers.Layer):
    def __init__(self, cfg: CaiTConfig, drop_path_rate: float, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        self.norm1 = norm_layer(name="norm1")
        self.attn = ClassAttention(
            embed_dim=cfg.embed_dim,
            nb_heads=cfg.nb_heads,
            qkv_bias=cfg.qkv_bias,
            attn_drop_rate=0.0,
            proj_drop_rate=0.0,
            name="attn",
        )
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp = MLP(
            hidden_dim=int(cfg.embed_dim * cfg.mlp_ratio),
            embed_dim=cfg.embed_dim,
            drop_rate=0.0,
            act_layer=cfg.act_layer,
            name="mlp",
        )

    def build(self, input_shape):
        self.gamma_1 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.cfg.init_scale),
            trainable=True,
            name="gamma_1",
        )
        self.gamma_2 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.cfg.init_scale),
            trainable=True,
            name="gamma_2",
        )

    def call(self, x, training=False):
        x_cls = tf.expand_dims(x[:, 0], axis=1)

        u = self.norm1(x, training=training)
        u = self.gamma_1 * self.attn(u)
        u = self.drop_path(u, training=training)
        x_cls = x_cls + u

        shortcut = x_cls
        x_cls = self.norm2(x_cls, training=training)
        x_cls = self.mlp(x_cls, training=training)
        x_cls = self.gamma_2 * x_cls
        x_cls = self.drop_path(x_cls, training=training)
        x_cls = x_cls + shortcut

        x = tf.concat((x_cls, x[:, 1:]), axis=1)
        return x


class TalkingHeadAttention(tf.keras.layers.Layer):
    def __init__(
        self,
        embed_dim: int,
        nb_heads: int,
        qkv_bias: bool,
        attn_drop_rate: float,
        proj_drop_rate: float,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.nb_heads = nb_heads
        self.scale = (embed_dim // nb_heads) ** -0.5

        self.qkv = tf.keras.layers.Dense(
            units=3 * embed_dim,
            use_bias=qkv_bias,
            name="qkv",
        )
        self.attn_drop = tf.keras.layers.Dropout(rate=attn_drop_rate)
        self.proj = tf.keras.layers.Dense(units=embed_dim, name="proj")
        self.proj_l = tf.keras.layers.Dense(units=nb_heads, name="proj_l")
        self.proj_w = tf.keras.layers.Dense(units=nb_heads, name="proj_w")
        self.proj_drop = tf.keras.layers.Dropout(rate=proj_drop_rate)

    def call(self, x, training=False):
        # B (batch size), N (sequence length), D (embedding dimension),
        # H (number of heads)
        batch_size, seq_length = tf.unstack(tf.shape(x)[:2])
        qkv = self.qkv(x)  # (B, N, 3*D)
        qkv = tf.reshape(qkv, (batch_size, seq_length, 3, self.nb_heads, -1))
        qkv = tf.transpose(qkv, (2, 0, 3, 1, 4))  # (3, B, H, N, D/H)
        q, k, v = qkv[0], qkv[1], qkv[2]
        q = self.scale * q

        attn = tf.linalg.matmul(q, k, transpose_b=True)  # (B, H, N, N)
        attn = tf.transpose(attn, (0, 2, 3, 1))  # (B, N, N, H)
        attn = self.proj_l(attn)  # (B, N, N, H)
        attn = tf.transpose(attn, (0, 3, 1, 2))  # (B, H, N, N)
        attn = tf.nn.softmax(attn, axis=-1)  # (B, H, N, N)
        attn = tf.transpose(attn, (0, 2, 3, 1))  # (B, N, N, H)
        attn = self.proj_w(attn)  # (B, N, N, H)
        attn = tf.transpose(attn, (0, 3, 1, 2))  # (B, H, N, N)
        attn = self.attn_drop(attn, training=training)

        x = tf.linalg.matmul(attn, v)  # (B, H, N, D/H)
        x = tf.transpose(x, (0, 2, 1, 3))  # (B, N, H, D/H)
        x = tf.reshape(x, (batch_size, seq_length, -1))  # (B, N, D)

        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return x


class LayerScaleBlock(tf.keras.layers.Layer):
    def __init__(self, cfg: CaiTConfig, drop_path_rate: float, **kwargs):
        super().__init__(**kwargs)
        self.cfg = cfg

        norm_layer = norm_layer_factory(cfg.norm_layer)
        self.norm1 = norm_layer(name="norm1")
        self.attn = TalkingHeadAttention(
            embed_dim=cfg.embed_dim,
            nb_heads=cfg.nb_heads,
            qkv_bias=cfg.qkv_bias,
            attn_drop_rate=cfg.attn_drop_rate,
            proj_drop_rate=cfg.drop_rate,
            name="attn",
        )
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.norm2 = norm_layer(name="norm2")
        self.mlp = MLP(
            hidden_dim=int(cfg.embed_dim * cfg.mlp_ratio),
            embed_dim=cfg.embed_dim,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp",
        )

    def build(self, input_shape):
        self.gamma_1 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.cfg.init_scale),
            trainable=True,
            name="gamma_1",
        )
        self.gamma_2 = self.add_weight(
            shape=(self.cfg.embed_dim,),
            initializer=tf.keras.initializers.Constant(value=self.cfg.init_scale),
            trainable=True,
            name="gamma_2",
        )

    def call(self, x, training=False):
        shortcut = x
        x = self.norm1(x, training=training)
        x = self.attn(x, training=training)
        x = self.gamma_1 * x
        x = self.drop_path(x, training=training)
        x = x + shortcut

        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = self.gamma_2 * x
        x = self.drop_path(x, training=training)
        x = x + shortcut
        return x


@keras_serializable
class CaiT(tf.keras.Model):
    cfg_class = CaiTConfig

    def __init__(self, cfg: CaiTConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg
        self.nb_features = cfg.embed_dim

        norm_layer = norm_layer_factory(cfg.norm_layer)

        self.patch_embed = PatchEmbeddings(
            patch_size=cfg.patch_size,
            embed_dim=cfg.embed_dim,
            norm_layer="",
            name="patch_embed",
        )
        self.pos_drop = tf.keras.layers.Dropout(rate=cfg.drop_rate)

        dpr = [cfg.drop_path_rate for _ in range(cfg.nb_blocks)]
        self.blocks = [
            LayerScaleBlock(cfg=cfg, drop_path_rate=rate, name=f"blocks/{j}")
            for j, rate in zip(range(cfg.nb_blocks), dpr)
        ]
        self.block_token_only = [
            LayerScaleBlockClassAttention(
                cfg=cfg, drop_path_rate=0.0, name=f"blocks_token_only/{j}"
            )
            for j in range(2)
        ]
        self.norm = norm_layer(name="norm")
        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    def build(self, input_shape):
        self.cls_token = self.add_weight(
            shape=(1, 1, self.cfg.embed_dim),
            initializer="zeros",
            trainable=True,
            name="cls_token",
        )
        self.pos_embed = self.add_weight(
            shape=(1, self.cfg.nb_patches, self.cfg.embed_dim),
            initializer="zeros",
            trainable=True,
            name="pos_embed",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        return (
            ["patch_embedding"]
            + [f"block_{j}" for j in range(self.cfg.nb_blocks)]
            + ["features_cls_token"]
            + [f"block_cls_token_{j}" for j in range(2)]
            + ["features_all", "features", "logits"]
        )

    def transform_pos_embed(self, src_weights, target_cfg: CaiTConfig):
        return interpolate_pos_embeddings(
            pos_embed=self.pos_embed,
            src_grid_size=self.cfg.grid_size,
            tgt_grid_size=target_cfg.grid_size,
            nb_tokens=0,  # In CaiT the class token is added only near the end
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x, grid_size = self.patch_embed(x, return_shape=True)
        if not self.cfg.interpolate_input:
            x = x + self.pos_embed
        else:
            pos_embed = interpolate_pos_embeddings(
                self.pos_embed,
                src_grid_size=self.cfg.grid_size,
                tgt_grid_size=grid_size,
                nb_tokens=0,  # In CaiT the class token is added only near the end
            )
            x = x + pos_embed
        x = self.pos_drop(x, training=training)
        features["patch_embedding"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x

        batch_size = tf.shape(x)[0]
        cls_token = tf.repeat(self.cls_token, repeats=batch_size, axis=0)
        x = tf.concat((cls_token, x), axis=1)
        features["features_cls_token"] = x

        for j, block in enumerate(self.block_token_only):
            x = block(x, training=training)
            features[f"block_cls_token_{j}"] = x

        x = self.norm(x, training=training)
        features["features_all"] = x
        x = x[:, 0]
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def cait_xxs24_224():
    cfg = CaiTConfig(
        name="cait_xxs24_224",
        url="[timm]",
        input_size=(224, 224),
        patch_size=16,
        embed_dim=192,
        nb_blocks=24,
        nb_heads=4,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_xxs24_384():
    cfg = CaiTConfig(
        name="cait_xxs24_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=192,
        nb_blocks=24,
        nb_heads=4,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_xxs36_224():
    cfg = CaiTConfig(
        name="cait_xxs36_224",
        url="[timm]",
        input_size=(224, 224),
        patch_size=16,
        embed_dim=192,
        nb_blocks=36,
        nb_heads=4,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_xxs36_384():
    cfg = CaiTConfig(
        name="cait_xxs36_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=192,
        nb_blocks=36,
        nb_heads=4,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_xs24_384():
    cfg = CaiTConfig(
        name="cait_xs24_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=288,
        nb_blocks=24,
        nb_heads=6,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_s24_224():
    cfg = CaiTConfig(
        name="cait_s24_224",
        url="[timm]",
        input_size=(224, 224),
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        nb_heads=8,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_s24_384():
    cfg = CaiTConfig(
        name="cait_s24_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=384,
        nb_blocks=24,
        nb_heads=8,
        init_scale=1e-5,
    )
    return CaiT, cfg


@register_model
def cait_s36_384():
    cfg = CaiTConfig(
        name="cait_s36_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=384,
        nb_blocks=36,
        nb_heads=8,
        init_scale=1e-6,
    )
    return CaiT, cfg


@register_model
def cait_m36_384():
    cfg = CaiTConfig(
        name="cait_m36_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=16,
        embed_dim=768,
        nb_blocks=36,
        nb_heads=16,
        init_scale=1e-6,
    )
    return CaiT, cfg


@register_model
def cait_m48_448():
    cfg = CaiTConfig(
        name="cait_m48_448",
        url="[timm]",
        input_size=(448, 448),
        patch_size=16,
        embed_dim=768,
        nb_blocks=48,
        nb_heads=16,
        init_scale=1e-6,
    )
    return CaiT, cfg



"""
TensorFlow implementation of the Swin Transformer

Based on the implementation by Rishigami
Source: https://github.com/rishigami/Swin-Transformer-TF

Paper: Swin Transformer: Hierarchical Vision Transformer using Shifted Windows
Link: https://arxiv.org/abs/2103.14030
Official implementation: https://github.com/microsoft/Swin-Transformer

Copyright: 2021 Martins Bruveris
Copyright: 2021 Rishigami
"""
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import MLP, DropPath, PatchEmbeddings, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# model_registry will add each entrypoint fn to this
__all__ = ["SwinTransformer", "SwinTransformerConfig"]


@dataclass
class SwinTransformerConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: int = 4
    embed_dim: int = 96
    nb_blocks: Tuple = (2, 2, 6, 2)
    nb_heads: Tuple = (3, 6, 12, 24)
    window_size: int = 7
    mlp_ratio: float = 4.0
    qkv_bias: bool = True
    # Regularization
    drop_rate: float = 0.0
    attn_drop_rate: float = 0.0
    drop_path_rate: float = 0.1
    # Other parameters
    norm_layer: str = "layer_norm"
    act_layer: str = "gelu"
    patch_norm: bool = True
    # Parameters for inference
    interpolate_input: bool = False
    crop_pct: float = 0.9
    interpolation: str = "bicubic"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    # Weight transfer
    first_conv: str = "patch_embed/proj"
    classifier: str = "head"

    @property
    def patch_resolution(self):
        """Resolution of grid of patches."""
        return (
            self.input_size[0] // self.patch_size,
            self.input_size[1] // self.patch_size,
        )

    @property
    def nb_patches(self):
        return self.patch_resolution[0] * self.patch_resolution[1]


def window_partition(x: tf.Tensor, window_size: int):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): Window size

    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    b, h, w, c = tf.unstack(tf.shape(x))
    x = tf.reshape(
        x, shape=(-1, h // window_size, window_size, w // window_size, window_size, c)
    )
    x = tf.transpose(x, perm=(0, 1, 3, 2, 4, 5))
    windows = tf.reshape(x, shape=(-1, window_size, window_size, c))
    return windows


def window_reverse(windows, window_size, h, w, c):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        h (int): Height of image
        w (int): Width of image
        c (int): Number of channels in image

    Returns:
        x: (B, H, W, C)
    """
    x = tf.reshape(
        windows,
        shape=(-1, h // window_size, w // window_size, window_size, window_size, c),
    )
    x = tf.transpose(x, perm=(0, 1, 3, 2, 4, 5))
    x = tf.reshape(x, shape=(-1, h, w, c))
    return x


class WindowAttention(tf.keras.layers.Layer):
    def __init__(
        self,
        cfg: SwinTransformerConfig,
        embed_dim: int,
        nb_heads: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cfg = cfg
        self.embed_dim = embed_dim
        self.nb_heads = nb_heads

        self.qkv = tf.keras.layers.Dense(
            embed_dim * 3, use_bias=cfg.qkv_bias, name="qkv"
        )
        self.attn_drop = tf.keras.layers.Dropout(cfg.attn_drop_rate)
        self.proj = tf.keras.layers.Dense(embed_dim, name="proj")
        self.proj_drop = tf.keras.layers.Dropout(cfg.drop_rate)

    def build(self, input_shape):
        window_size = self.cfg.window_size

        # The weights have to be created inside the build() function for the right
        # name scope to be set.
        self.relative_position_bias_table = self.add_weight(
            name="relative_position_bias_table",
            shape=((2 * window_size - 1) * (2 * window_size - 1), self.nb_heads),
            initializer=tf.initializers.Zeros(),
            trainable=True,
        )

        coords_h = np.arange(window_size)
        coords_w = np.arange(window_size)
        coords = np.stack(np.meshgrid(coords_h, coords_w, indexing="ij"))
        coords_flatten = coords.reshape(2, -1)
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]
        relative_coords = relative_coords.transpose((1, 2, 0))
        relative_coords[:, :, 0] += window_size - 1
        relative_coords[:, :, 1] += window_size - 1
        relative_coords[:, :, 0] *= 2 * window_size - 1
        relative_position_index = relative_coords.sum(-1).astype(np.int64)
        self.relative_position_index = tf.Variable(
            name="relative_position_index",
            initial_value=tf.convert_to_tensor(relative_position_index),
            trainable=False,
        )

    def call(self, inputs, training=False):
        nb_heads = self.nb_heads
        window_size = self.cfg.window_size

        # Inputs are the batch and the attention mask
        x, mask = inputs[0], inputs[1]
        _, n, c = tf.unstack(tf.shape(x))

        qkv = self.qkv(x)
        qkv = tf.reshape(qkv, shape=(-1, n, 3, nb_heads, c // nb_heads))
        qkv = tf.transpose(qkv, perm=(2, 0, 3, 1, 4))
        q, k, v = tf.unstack(qkv)

        scale = (self.embed_dim // nb_heads) ** -0.5
        q = q * scale
        attn = q @ tf.transpose(k, perm=(0, 1, 3, 2))
        relative_position_bias = tf.gather(
            self.relative_position_bias_table,
            tf.reshape(self.relative_position_index, shape=(-1,)),
        )
        relative_position_bias = tf.reshape(
            relative_position_bias,
            shape=(window_size**2, window_size**2, -1),
        )
        relative_position_bias = tf.transpose(relative_position_bias, perm=(2, 0, 1))
        attn = attn + tf.expand_dims(relative_position_bias, axis=0)

        nw = mask.get_shape()[0]  # tf.shape(mask)[0]
        mask = tf.expand_dims(tf.expand_dims(mask, axis=1), axis=0)
        mask = tf.cast(mask, attn.dtype)
        attn = tf.reshape(attn, shape=(-1, nw, nb_heads, n, n)) + mask
        attn = tf.reshape(attn, shape=(-1, nb_heads, n, n))
        attn = tf.nn.softmax(attn, axis=-1)
        attn = self.attn_drop(attn, training=training)

        x = tf.transpose((attn @ v), perm=(0, 2, 1, 3))
        x = tf.reshape(x, shape=(-1, n, c))
        x = self.proj(x)
        x = self.proj_drop(x, training=training)
        return x


class SwinTransformerBlock(tf.keras.layers.Layer):
    def __init__(
        self,
        cfg: SwinTransformerConfig,
        input_size: Tuple[int, int],
        embed_dim: int,
        nb_heads: int,
        drop_path_rate: float,
        shift_size: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cfg = cfg
        self.input_size = input_size
        self.shift_size = shift_size
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.window_size = cfg.window_size

        # If the image resolution is smaller than the window size, there is no point
        # shifting windows, since we already capture the global context in that case.
        if min(self.input_size) <= self.window_size:
            self.shift_size = 0
            self.window_size = min(self.input_size)

        self.norm1 = self.norm_layer(name="norm1")
        self.attn = WindowAttention(
            cfg=cfg,
            embed_dim=embed_dim,
            nb_heads=nb_heads,
            name="attn",
        )
        self.attn_mask = None
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.norm2 = self.norm_layer(name="norm2")
        self.mlp = MLP(
            hidden_dim=int(embed_dim * cfg.mlp_ratio),
            embed_dim=embed_dim,
            drop_rate=cfg.drop_rate,
            act_layer=cfg.act_layer,
            name="mlp",
        )

    def build(self, input_shape):
        h, w = self.input_size
        window_size = self.window_size
        shift_size = self.shift_size

        if shift_size > 0:
            img_mask = np.zeros([1, h, w, 1])
            h_slices = (
                slice(0, -window_size),
                slice(-window_size, -shift_size),
                slice(-shift_size, None),
            )
            w_slices = (
                slice(0, -window_size),
                slice(-window_size, -shift_size),
                slice(-shift_size, None),
            )
            cnt = 0
            for h_slice in h_slices:
                for w_slice in w_slices:
                    img_mask[:, h_slice, w_slice, :] = cnt
                    cnt += 1

            img_mask = tf.convert_to_tensor(img_mask)
            mask_windows = window_partition(img_mask, window_size)
            mask_windows = tf.reshape(mask_windows, shape=(-1, window_size**2))
            attn_mask = tf.expand_dims(mask_windows, axis=1) - tf.expand_dims(
                mask_windows, axis=2
            )
            attn_mask = tf.where(attn_mask != 0, -100.0, attn_mask)
            attn_mask = tf.where(attn_mask == 0, 0.0, attn_mask)
            # Only the non-trivial attention mask is a Variable, because in the PyTorch
            # model, only the non-trivial attention mask exists. The trivial maks
            # is replaced by None and if-statements in the call function.
            self.attn_mask = tf.Variable(
                initial_value=attn_mask, trainable=False, name="attn_mask"
            )
        else:
            # Attention mask is applied additively, so zero-mask has no effect
            # Broadcasting will take care of mapping it to the correct dimensions.
            self.attn_mask = tf.Variable(
                initial_value=tf.zeros((1,)), trainable=False, name="attn_mask"
            )

    def call(self, x, training=False):
        window_size = self.window_size
        shift_size = self.shift_size

        h, w = self.input_size
        b, l, c = tf.unstack(tf.shape(x))

        shortcut = x
        x = self.norm1(x, training=training)
        x = tf.reshape(x, shape=(-1, h, w, c))

        # Cyclic shift (Identify, if shift_size == 0)
        shifted_x = tf.roll(x, shift=(-shift_size, -shift_size), axis=[1, 2])

        # Partition windows
        x_windows = window_partition(shifted_x, window_size)
        x_windows = tf.reshape(x_windows, shape=(-1, window_size**2, c))

        # W-MSA/SW-MSA
        attn_windows = self.attn([x_windows, self.attn_mask])

        # Merge windows
        attn_windows = tf.reshape(attn_windows, shape=(-1, window_size, window_size, c))
        shifted_x = window_reverse(attn_windows, window_size, h, w, c)

        # Reverse cyclic shift
        x = tf.roll(shifted_x, shift=(shift_size, shift_size), axis=(1, 2))
        x = tf.reshape(x, shape=[-1, h * w, c])

        # Residual connection
        x = self.drop_path(x, training=training)
        x = x + shortcut

        # MLP
        shortcut = x
        x = self.norm2(x, training=training)
        x = self.mlp(x, training=training)
        x = self.drop_path(x, training=training)
        x = x + shortcut

        return x


class PatchMerging(tf.keras.layers.Layer):
    def __init__(
        self,
        cfg: SwinTransformerConfig,
        input_size: Tuple[int, int],
        embed_dim: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cfg = cfg
        self.input_size = input_size
        self.norm_layer = norm_layer_factory(cfg.norm_layer)

        self.reduction = tf.keras.layers.Dense(
            units=2 * embed_dim, use_bias=False, name="reduction"
        )
        self.norm = self.norm_layer(name="norm")

    def call(self, x, training=False):
        h, w = self.input_size
        b, l, c = tf.unstack(tf.shape(x))

        x = tf.reshape(x, shape=(-1, h, w, c))
        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = tf.concat((x0, x1, x2, x3), axis=-1)
        x = tf.reshape(x, shape=(-1, (h // 2) * (w // 2), 4 * c))

        x = self.norm(x, training=training)
        x = self.reduction(x)
        return x


class SwinTransformerStage(tf.keras.layers.Layer):
    def __init__(
        self,
        cfg: SwinTransformerConfig,
        input_size: Tuple[int, int],
        embed_dim: int,
        nb_blocks: int,
        nb_heads: int,
        drop_path_rate: np.ndarray,
        downsample: bool,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cfg = cfg

        self.blocks = [
            SwinTransformerBlock(
                cfg=cfg,
                input_size=input_size,
                embed_dim=embed_dim,
                nb_heads=nb_heads,
                drop_path_rate=drop_path_rate[idx],
                shift_size=0 if idx % 2 == 0 else cfg.window_size // 2,
                name=f"blocks/{idx}",
            )
            for idx in range(nb_blocks)
        ]
        if downsample:
            self.downsample = PatchMerging(
                cfg=cfg, input_size=input_size, embed_dim=embed_dim, name="downsample"
            )
        else:
            self.downsample = tf.keras.layers.Activation("linear")

    def call(self, x, training=False, return_features=False):
        features = {}
        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x

        x = self.downsample(x, training=training)
        features["features"] = x
        return (x, features) if return_features else x


@keras_serializable
class SwinTransformer(tf.keras.Model):
    cfg_class = SwinTransformerConfig

    def __init__(self, cfg: SwinTransformerConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg
        self.norm_layer = norm_layer_factory(cfg.norm_layer)

        # Split image into non-overlapping patches
        self.patch_embed = PatchEmbeddings(
            patch_size=cfg.patch_size,
            embed_dim=cfg.embed_dim,
            norm_layer=cfg.norm_layer,
            name="patch_embed",
        )
        self.drop = tf.keras.layers.Dropout(cfg.drop_rate)

        # Stochastic depth
        dpr = np.linspace(0.0, cfg.drop_path_rate, sum(cfg.nb_blocks))

        # Build stages
        self.stages = []
        nb_stages = len(cfg.nb_blocks)
        block_idx_to = 0
        for idx in range(nb_stages):
            block_idx_from = block_idx_to
            block_idx_to = block_idx_to + cfg.nb_blocks[idx]

            self.stages.append(
                SwinTransformerStage(
                    cfg=cfg,
                    input_size=(
                        cfg.patch_resolution[0] // (2**idx),
                        cfg.patch_resolution[1] // (2**idx),
                    ),
                    embed_dim=int(cfg.embed_dim * 2**idx),
                    nb_blocks=cfg.nb_blocks[idx],
                    nb_heads=cfg.nb_heads[idx],
                    drop_path_rate=dpr[block_idx_from:block_idx_to],
                    downsample=idx < nb_stages - 1,  # Don't downsample the last stage
                    name=f"layers/{idx}",
                )
            )
        self.norm = self.norm_layer(name="norm")
        self.pool = tf.keras.layers.GlobalAveragePooling1D()
        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")  # Identity layer
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        names = ["patch_embedding"]
        k = 0
        for j in range(len(self.cfg.nb_blocks)):
            for _ in range(self.cfg.nb_blocks[j]):
                names.append(f"block_{k}")
                k += 1
            names.append(f"stage_{j}")
        names += ["features_all", "features", "logits"]
        return names

    @property
    def keys_to_ignore_on_load_missing(self) -> List[str]:
        names = []
        for j in range(len(self.cfg.nb_blocks)):
            for k in range(self.cfg.nb_blocks[j]):
                names.append(f"layers/{j}/blocks/{k}/attn_mask")
                names.append(f"layers/{j}/blocks/{k}/attn/relative_position_index")
        return names

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x = self.patch_embed(x, training=training)
        x = self.drop(x, training=training)
        features["patch_embedding"] = x

        block_idx = 0
        for stage_idx, stage in enumerate(self.stages):
            x = stage(x, training=training, return_features=return_features)
            if return_features:
                x, stage_features = x
                for k in range(self.cfg.nb_blocks[stage_idx]):
                    features[f"block_{block_idx}"] = stage_features[f"block_{k}"]
                    block_idx += 1
                features[f"stage_{stage_idx}"] = stage_features["features"]

        x = self.norm(x, training=training)
        features["features_all"] = x
        x = self.pool(x)
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def swin_tiny_patch4_window7_224():
    """Swin-T @ 224x224, trained ImageNet-1k"""
    cfg = SwinTransformerConfig(
        name="swin_tiny_patch4_window7_224",
        url="[timm]",
        patch_size=4,
        embed_dim=96,
        nb_blocks=(2, 2, 6, 2),
        nb_heads=(3, 6, 12, 24),
        window_size=7,
    )
    return SwinTransformer, cfg


@register_model
def swin_small_patch4_window7_224():
    """Swin-S @ 224x224, trained ImageNet-1k"""
    cfg = SwinTransformerConfig(
        name="swin_small_patch4_window7_224",
        url="[timm]",
        patch_size=4,
        embed_dim=96,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(3, 6, 12, 24),
        window_size=7,
    )
    return SwinTransformer, cfg


@register_model
def swin_base_patch4_window12_384():
    """Swin-B @ 384x384, pretrained ImageNet-22k, fine tune 1k"""
    cfg = SwinTransformerConfig(
        name="swin_base_patch4_window12_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=4,
        embed_dim=128,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(4, 8, 16, 32),
        window_size=12,
        crop_pct=1.0,
    )
    return SwinTransformer, cfg


@register_model
def swin_base_patch4_window7_224():
    """Swin-B @ 224x224, pretrained ImageNet-22k, fine tune 1k"""
    cfg = SwinTransformerConfig(
        name="swin_base_patch4_window7_224",
        url="[timm]",
        patch_size=4,
        embed_dim=128,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(4, 8, 16, 32),
        window_size=7,
    )
    return SwinTransformer, cfg


@register_model
def swin_base_patch4_window12_384_in22k():
    """Swin-B @ 384x384, trained ImageNet-22k"""
    cfg = SwinTransformerConfig(
        name="swin_base_patch4_window12_384_in22k",
        url="[timm]",
        nb_classes=21841,
        input_size=(384, 384),
        patch_size=4,
        embed_dim=128,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(4, 8, 16, 32),
        window_size=12,
        crop_pct=1.0,
    )
    return SwinTransformer, cfg


@register_model
def swin_base_patch4_window7_224_in22k():
    """Swin-B @ 224x224, trained ImageNet-22k"""
    cfg = SwinTransformerConfig(
        name="swin_base_patch4_window7_224_in22k",
        url="[timm]",
        nb_classes=21841,
        patch_size=4,
        embed_dim=128,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(4, 8, 16, 32),
        window_size=7,
    )
    return SwinTransformer, cfg


@register_model
def swin_large_patch4_window12_384():
    """Swin-L @ 384x384, pretrained ImageNet-22k, fine tune 1k"""
    cfg = SwinTransformerConfig(
        name="swin_large_patch4_window12_384",
        url="[timm]",
        input_size=(384, 384),
        patch_size=4,
        embed_dim=192,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(6, 12, 24, 48),
        window_size=12,
        crop_pct=1.0,
    )
    return SwinTransformer, cfg


@register_model
def swin_large_patch4_window7_224():
    """Swin-L @ 224x224, pretrained ImageNet-22k, fine tune 1k"""
    cfg = SwinTransformerConfig(
        name="swin_large_patch4_window7_224",
        url="[timm]",
        patch_size=4,
        embed_dim=192,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(6, 12, 24, 48),
        window_size=7,
    )
    return SwinTransformer, cfg


@register_model
def swin_large_patch4_window12_384_in22k():
    """Swin-L @ 384x384, trained ImageNet-22k"""
    cfg = SwinTransformerConfig(
        name="swin_large_patch4_window12_384_in22k",
        url="[timm]",
        nb_classes=21841,
        input_size=(384, 384),
        patch_size=4,
        embed_dim=192,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(6, 12, 24, 48),
        window_size=12,
        crop_pct=1.0,
    )
    return SwinTransformer, cfg


@register_model
def swin_large_patch4_window7_224_in22k():
    """Swin-L @ 224x224, trained ImageNet-22k"""
    cfg = SwinTransformerConfig(
        name="swin_large_patch4_window7_224_in22k",
        url="[timm]",
        nb_classes=21841,
        patch_size=4,
        embed_dim=192,
        nb_blocks=(2, 2, 18, 2),
        nb_heads=(6, 12, 24, 48),
        window_size=7,
    )
    return SwinTransformer, cfg



from .cache import (  # noqa: F401
    cached_model_path,
    clear_model_cache,
    get_dir,
    list_cached_models,
    set_dir,
    set_model_cache,
)
from .constants import *  # noqa: F401
from .etc import make_divisible, to_2tuple  # noqa: F401
from .timm import load_pth_url_weights, load_timm_weights  # noqa: F401



# These imports have to be absolute because this file is being executed as a script
import tfimm.train.datasets  # noqa: F401
from tfimm.train import run, setup_logging


def main():
    run(cfg={}, parse_args=True)


if __name__ == "__main__":
    setup_logging()
    main()



# TODO: Test notebook with Colab
# TODO: Add "Open in Colab" badge to notebook (see SAM)
# TODO: Test mixed precision behaviour
# TODO: Convert PT models to TF and upload to GitHub
import math
from typing import Callable, Optional, Tuple

import numpy as np
import tensorflow as tf

from tfimm.models.factory import create_preprocessing

from .sam import SegmentAnythingModel


class SAMPredictor:
    """
    User-friendly interface to the Segment Anything model. Uses SAM to calculate the
    image embedding for an image, and then allows repeated, efficient mask prediction
    given prompts.

    While internally TF is used for inference, the inputs and return values in this
    class are numpy arrays for ease of use.

    Args:
        model: The model used for mask prediction.
        preprocessing: Preprocessing function for the model. If not provided we
            will query ``tfimm`` using the model name.
    """

    def __init__(
        self, model: SegmentAnythingModel, preprocessing: Optional[Callable] = None
    ):
        if preprocessing is None:
            preprocessing = create_preprocessing(
                model.cfg.name, in_channels=model.cfg.in_channels, dtype=tf.float32
            )

        self.model = model
        self.preprocessing = preprocessing

        # These attributes are set when calling `set_image()`.
        self.resizer: Optional[ImageResizer] = None
        self.image_embedding = None
        self.image_set = False

    def set_image(self, image: np.ndarray):
        """
        Calculates and stores the image embeddings for the provided image, allowing
        masks to be predicted much faster.

        Args:
            image: An array of shape (H, W, C) with pixel values in [0, 255]. The image
                can be any shape, and it will be resized and padded to the model input
                shape as necessary.

        Returns:
            Nothing. The image embedding and resizing information are stored in the
            class.
        """
        if self.model.cfg.fixed_input_size:
            self.resizer = ImageResizer(
                src_size=image.shape[:2], dst_size=self.model.cfg.input_size
            )
        else:
            # If the model allows flexible input sizes, we simply pad the image to
            # the nearest multiple of patch_size.
            patch_size = self.model.cfg.encoder_patch_size
            dst_size = (
                patch_size * math.ceil(image.shape[0] / patch_size),
                patch_size * math.ceil(image.shape[1] / patch_size),
            )
            self.resizer = ImageResizer(
                src_size=image.shape[:2], dst_size=dst_size, pad_only=True
            )

        # We scale the image to the largest size possible that fits input_size.
        image = self.resizer.scale_image(image)
        image = self.resizer.pad_image(image)  # Pad to input_size
        image = image[np.newaxis, ...]  # Add batch dimension

        # Now we move to TF world and compute embeddings
        image = tf.convert_to_tensor(image)
        image = self.preprocessing(image)
        self.image_embedding = self.model.image_encoder(image, training=False)

        self.image_set = True

    def clear_image(self):
        """Unsets the image and forgets the embedding."""
        self.resizer = None
        self.image_embedding = None
        self.image_set = False

    def input_size(self):
        """Returns the input size to the model."""
        if self.image_set:
            return self.resizer.dst_size
        elif self.model.cfg.fixed_input_size:
            return self.model.cfg.input_size
        else:
            raise ValueError(
                "To determine model input size need to set image or use a model with "
                "a fixed input size."
            )

    def mask_size(self):
        """Returns the mask prompt input size to the model."""
        return self.model.mask_size(self.input_size())

    def preprocess_masks(self, mask: np.ndarray) -> np.ndarray:
        """
        Preprocesses a mask from the pixel space of the original image (H0, W0), to the
        correct input size to the model. Note that the mask should be a mask of
        logits and *not* the thresholded version.

        Args:
            mask: An array of shape (M, H0, W0) or (N, M, H0, W0), where (H0, W0) is the
                original size of the input image.

        Returns:
            Preprocessed mask of shape (M, H', W') or (N, M, H', W') as given by
                ``mask_size``.
        """
        # First we convert mask to full mask at input_size resolution
        mask = self.resizer.scale_image(mask, channels_last=False)
        mask = self.resizer.pad_image(mask, channels_last=False)

        # Then we rescale to mask_size
        mask_size = self.mask_size()
        mask = self.resizer.scale_to_size(mask, size=mask_size, channels_last=False)
        return mask

    def __call__(
        self,
        points: Optional[np.ndarray] = None,
        labels: Optional[np.ndarray] = None,
        boxes: Optional[np.ndarray] = None,
        masks: Optional[np.ndarray] = None,
        multimask_output: bool = True,
        return_logits: bool = False,
    ):
        """
        Predicts masks end-to-end for the given prompts. We assume that the image has
        already been set.

        The original image size is (H0, W0). After resizing and padding the image size
        becomes (H, W) given by ``input_size`` (usually (1024, 1024)). Mask input and
        logit output will have shape (H', W') given by ``mask_size`` (usually H'=H/4).

        One can use ``preprocess_masks`` to transform an input mask from (H0, W0) to
        (H', W').

        Prompts can also be batched, i.e., have the shape (N, M1, 2) for points;
        (N, M1) for point labels; (N, M2, 4) for boxes; and (N, M3, H', W') for mask
        prompts. Note that in this case we number and type of prompts is the same for
        each batch element. The return values will have the same batch dimension, i.e.,
        (N, K, H, W) for predicted masks, etc.

        Args:
            points: An (M1, 2) array of point prompts with coordinates in pixel
                space of the original image (H0, W0).
            labels: An (M1,) array of labels for point prompts. 1 indicates a
                foreground point and 0 indicates a background point.
            boxes: An (M2, 4) tensor of box prompts of form (left, top, right,
                bottom) with coordinates in pixel space of the original image (H0, W0).
            masks: An (M3, H', W') tensor of mask inputs, where (H', W') is the mask
                size (usually H'=H/4).
            multimask_output: If True, we return multiple nested masks for each prompt.
            return_logits: If True, we don't threshold the upscaled mask.

        Returns:
            * Masks, an (K, H, W) bool array of binary masked predictions, where K is
              determined by the multimask_output parameter. It is either 1, if
              ``multimask_output=False`` or given by the ``nb_multimask_outputs``
              parameter in the model configuration.
            * Scores, an (K,) array with the model's predictions of mask quality.
            * Logits, an (K, H', W') array with low resoulution logits, where usually
              H'=H/4 and W'=W/4. This can be passed as mask input to subsequent
              iterations of prediction.
        """
        if not self.image_set:
            raise ValueError("Need to set image before calling predict().")

        points = np.asarray(points) if points is not None else None
        labels = np.asarray(labels) if labels is not None else None
        boxes = np.asarray(boxes) if boxes is not None else None
        masks = np.asarray(masks) if masks is not None else None

        batch_shape = self._batch_shape(points, labels, boxes, masks)

        if points is None:
            points = np.zeros(batch_shape + (0, 2), dtype=np.float32)
        if labels is None:
            labels = np.zeros(batch_shape + (0,), dtype=np.int32)
        if boxes is None:
            boxes = np.zeros(batch_shape + (0, 4), dtype=np.float32)
        if masks is None:
            mask_size = self.mask_size()
            masks = np.zeros(batch_shape + (0, *mask_size), dtype=np.float32)

        # Check that batch shapes are compatible
        if (
            points.shape[:-2] != batch_shape
            or labels.shape[:-1] != batch_shape
            or boxes.shape[:-2] != batch_shape
            or masks.shape[:-3] != batch_shape
        ):
            raise ValueError("All prompts must have the same batch shape.")

        # Add batch dimension if needed
        if batch_shape == ():
            points = points[np.newaxis]
            labels = labels[np.newaxis]
            boxes = boxes[np.newaxis]
            masks = masks[np.newaxis]

        points = self.resizer.scale_points(points)
        boxes = self.resizer.scale_boxes(boxes)

        points = tf.convert_to_tensor(points)
        labels = tf.convert_to_tensor(labels)
        boxes = tf.convert_to_tensor(boxes)
        masks = tf.convert_to_tensor(masks)

        masks, scores, logits = self._predict_tf(
            points, labels, boxes, masks, multimask_output
        )

        masks = masks.numpy()
        scores = scores.numpy()
        logits = logits.numpy()

        # Transform masks back to image size.
        masks = self.resizer.postprocess_mask(masks)

        # Remove batch dimension, if input didn't have it.
        if batch_shape == ():
            masks = masks[0]
            scores = scores[0]
            logits = logits[0]

        # We apply the threshold only at the very end
        if not return_logits:
            masks = masks > self.model.mask_threshold

        return masks, scores, logits

    def _predict_tf(self, points, labels, boxes, masks, multimask_output):
        n = tf.shape(points)[0]
        image_embedding = tf.tile(self.image_embedding, (n, 1, 1, 1))

        sparse_embeddings, dense_embeddings = self.model.prompt_encoder(
            inputs={"points": points, "labels": labels, "boxes": boxes, "masks": masks},
            training=False,
        )

        logits, scores = self.model.mask_decoder(
            inputs={
                "image_embeddings": image_embedding,
                "image_pe": self.model.get_image_pe(image_embedding),
                "sparse_embeddings": sparse_embeddings,
                "dense_embeddings": dense_embeddings,
            },
            training=False,
            multimask_output=multimask_output,
        )

        masks = self.model.postprocess_logits(
            logits, input_size=self.input_size(), return_logits=True
        )
        return masks, scores, logits

    @staticmethod
    def _batch_shape(points, labels, boxes, masks):
        """Returns (), if there is no batch dimension and (n,) if there is."""
        if points is not None:
            return points.shape[:-2]
        elif labels is not None:
            return labels.shape[:-1]
        elif boxes is not None:
            return boxes.shape[:-2]
        elif masks is not None:
            return masks.shape[:-3]
        else:
            return ()


class ImageResizer:
    """
    Utility class to resize images to the largest side that fits in a given shape while
    preserving the aspect ratio. It also provides methods to resize coordinates and
    bounding boxes and pad images.

    Args:
        src_size: Size of image before resizing. The resize object is image
            specific, i.e., for each source image size it is recommended to create
            a new ``ImageResizer`` object.
        dst_size: The target size after resizing (and padding).
        pad_only: If True, we don't do any resizing and only pad the image to
            ``dst_size``.
    """

    def __init__(
        self,
        src_size: Tuple[int, int],
        dst_size: Tuple[int, int],
        pad_only: bool = False,
    ):
        self.src_size = src_size
        self.dst_size = dst_size
        self.pad_only = pad_only

        self.scale, self.rescaled_size = self._get_scale()

    def _get_scale(self) -> Tuple[float, Tuple[int, int]]:
        """Calculate rescaling parameters."""
        if self.pad_only:
            # If we only pad, then scale is 1 and the rescaled size equal input size.
            return 1.0, self.src_size

        h_scale = self.dst_size[0] / self.src_size[0]
        w_scale = self.dst_size[1] / self.src_size[1]

        if h_scale >= w_scale:
            scale = w_scale
            rescaled_size = (int(scale * self.src_size[0]), self.dst_size[1])
        else:
            scale = h_scale
            rescaled_size = (self.dst_size[0], int(scale * self.src_size[1]))

        # This is just to avoid rounding issues. We make sure not to exceed dst_size.
        rescaled_size = (
            min(rescaled_size[0], self.dst_size[0]),
            min(rescaled_size[1], self.dst_size[1]),
        )
        return scale, rescaled_size

    @staticmethod
    def scale_to_size(
        image: np.ndarray,
        size: Tuple[int, int],
        channels_last: bool,
    ) -> np.ndarray:
        """
        Scales an image to a given size. In this method we ignore ``dst_size`` and do
        not attempt to preserve the aspect ratio.

        Args:
            image: Image to be resized. Can be a 3D array (H, W, C) or a 4D array
                (N, H, W, C). We also accept channels first ordering (used for
                segmentation masks, i.e., (C, H, W) or (N, C, H, W).
            size: Target size.
            channels_last: If True, images are in HWC format and if False in CHW format.

        Returns:
            Resized image array.
        """
        no_batch_dim = image.ndim == 3  # Image without batch dimension
        if no_batch_dim:
            image = image[np.newaxis]  # Add batch dimension
        if not channels_last:  # TF resize expects HWC format.
            image = np.transpose(image, (0, 2, 3, 1))

        # TF resize converts everything to floats, so we remember the dtype
        dtype = image.dtype

        # Why are we using TF here and not OpenCV? Because TFIMM does not have OpenCV
        # as a dependency and I don't want to introduce it just because of the one
        # resizing operation.
        image = tf.convert_to_tensor(image)
        image = tf.image.resize(image, size=size, method=tf.image.ResizeMethod.AREA)
        image = image.numpy().astype(dtype)

        if not channels_last:
            image = np.transpose(image, (0, 3, 1, 2))
        if no_batch_dim:
            image = image[0]

        return image

    def scale_image(self, image: np.ndarray, channels_last: bool = True) -> np.ndarray:
        """
        Applies aspect-ratio preserving scaling to an image.

        Args:
            image: Image to be resized. Can be a 3D array (H, W, C) or a 4D array
                (N, H, W, C). We also accept channels first ordering (used for
                segmentation masks, i.e., (C, H, W) or (N, C, H, W).
            channels_last: If True, images are in HWC format and if False in CHW format.

        Returns:
            Resized image with spatial dimensions given by ``rescaled_size``. The
            longest edge of the image will be equal to ``dst_size``.
        """
        return self.scale_to_size(image, self.rescaled_size, channels_last)

    def unscale_image(
        self, image: np.ndarray, channels_last: bool = True
    ) -> np.ndarray:
        """
        Reverses the scaling operation.

        Args:
            image: Image to be rescaled back to original size given by ``src_size``. We
                assume that image has size ``rescaled_size``, otherwise aspect ratio
                will not be preserved. Image can be 3D or 4D with channels before or
                after spatial dimension.
            channels_last: If True, images are in HWC format and if False in CHW format.

        Returns:
            Resized image with size ``src_size``.
        """
        return self.scale_to_size(image, self.src_size, channels_last)

    def pad_image(self, image: np.ndarray, channels_last: bool = True) -> np.ndarray:
        """
        Apply zero padding to an image to size ``dst_size``.

        Args:
            image: Image to be padded. Can be 3D or 4D tensor with channel dimension
                before or after spatial dimensions.
            channels_last: If True, images are in HWC format and if False in CHW format.

        Returns:
            Zero padded image of size ``dst_size``.
        """
        no_batch_dim = image.ndim == 3  # Image without batch dimension
        if no_batch_dim:
            image = image[np.newaxis]  # Add batch dimension
        if not channels_last:  # TF resize expects HWC format.
            image = np.transpose(image, (0, 2, 3, 1))

        # Pad shorter edge to model input size.
        pad_h = self.dst_size[0] - image.shape[1]
        pad_w = self.dst_size[1] - image.shape[2]
        if pad_h < 0 or pad_w < 0:
            raise ValueError("Cannot pad an image that is larger than dst_size.")

        image = np.pad(image, [[0, 0], [0, pad_h], [0, pad_w], [0, 0]])

        if not channels_last:
            image = np.transpose(image, (0, 3, 1, 2))
        if no_batch_dim:
            image = image[0]

        return image

    def scale_points(self, points: np.ndarray) -> np.ndarray:
        """
        Scale points by the same factor as the image.

        Args:
            points: Points to be scaled.

        Returns:
            Scaled points.
        """
        return self.scale * points

    def scale_boxes(self, boxes: np.ndarray) -> np.ndarray:
        """
        Scale bounding boxes by the same factor as the image.

        Args:
            boxes: Boxes to be scaled.

        Returns:
            Scaled boxes.
        """
        return self.scale * boxes

    def postprocess_mask(
        self, mask: np.ndarray, threshold: Optional[float] = None
    ) -> np.ndarray:
        """
        Convert an upscaled segmentation mask from ``dst_size`` back to ``src_size``
        by removing padding and unscaling.

        Args:
            mask: Segmentation mask, i.e., image with channels_first ordering, of size
                ``dst_size``. Should be a logit-mask, i.e., before thresholding.
            threshold: Optionally, we can apply thresholding after resizing to obtain
                a boolean mask.

        Returns:
            Mask of size ``src_size``.
        """
        mask = mask[..., : self.rescaled_size[0], : self.rescaled_size[1]]
        mask = self.unscale_image(mask, channels_last=False)
        if threshold is not None:
            mask = mask > threshold
        return mask



"""
Pre-Activation ResNet v2 with GroupNorm and Weight Standardization.

A TensorFlow implementation of ResNetV2 adapted from the Google Big-Transfoer (BiT)
source code at https://github.com/google-research/big_transfer to match `tfimm`
interfaces. The BiT weights have been included here as pretrained models from the `timm`
versions.

Additionally, supports non pre-activation bottleneck for use as a backbone for
Vision Transfomers (ViT) and extra padding support to allow porting of official
Hybrid ResNet pretrained weights from
https://github.com/google-research/vision_transformer

Thanks to the Google team for the above two repositories and associated papers:
* Big Transfer (BiT): General Visual Representation Learning
  Link: https://arxiv.org/abs/1912.11370
* An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale
  Link: https://arxiv.org/abs/2010.11929
* Knowledge distillation: A good teacher is patient and consistent
  Link: https://arxiv.org/abs/2106.05237

Copyright 2021 Martins Bruveris
Copyright 2020 Ross Wightman
Copyright 2020 Google LLC
"""
from dataclasses import dataclass
from typing import List, Tuple

import numpy as np
import tensorflow as tf

from tfimm.layers import (
    ClassifierHead,
    DropPath,
    StdConv2D,
    act_layer_factory,
    norm_layer_factory,
)
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_INCEPTION_MEAN, IMAGENET_INCEPTION_STD

# model_registry will add each entrypoint fn to this
__all__ = ["ResNetV2", "ResNetV2Config", "ResNetV2Stem"]


@dataclass
class ResNetV2Config(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    # Residual blocks
    nb_blocks: Tuple = (2, 2, 2, 2)
    nb_channels: Tuple = (256, 512, 1024, 2048)
    width_factor: int = 1
    preact: bool = True  # Preactivation structure for res blocks
    # Stem
    stem_width: int = 64
    stem_type: str = "fixed"
    # Pool
    global_pool: str = "avg"
    # Other params
    conv_padding: str = "symmetric"
    act_layer: str = "relu"
    norm_layer: str = "group_norm"
    # Regularization
    drop_rate: float = 0.0
    drop_path_rate: float = 0.0
    # Parameters for inference
    pool_size: int = 7  # For test-time pooling (not implemented yet)
    crop_pct: float = 0.875
    interpolation: str = "bilinear"
    # Preprocessing
    mean: Tuple[float, float, float] = IMAGENET_INCEPTION_MEAN
    std: Tuple[float, float, float] = IMAGENET_INCEPTION_STD
    # Weight transfer
    first_conv: str = "stem/conv"
    classifier: str = "head/fc"


def make_divisible(v, divisor=8):
    min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class PreActBottleneck(tf.keras.layers.Layer):
    """
    Pre-activation (v2) bottleneck block.

    Follows the implementation of "Identity Mappings in Deep Residual Networks":
    https://github.com/KaimingHe/resnet-1k-layers/blob/master/resnet-pre-act.lua

    Except it puts the stride on 3x3 conv when available.
    """

    def __init__(
        self,
        nb_channels,
        strides: int,
        downsample: bool,
        conv_padding: str,
        act_layer: str,
        norm_layer: str,
        drop_path_rate: float,
        bottleneck_ratio=0.25,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.act_layer = act_layer_factory(act_layer)
        self.norm_layer = norm_layer_factory(norm_layer)
        mid_channels = make_divisible(nb_channels * bottleneck_ratio)

        if downsample:
            self.downsample = DownsampleConv(
                nb_channels=nb_channels,
                strides=strides,
                preact=True,
                conv_padding=conv_padding,
                norm_layer=norm_layer,
                name="downsample",
            )
        else:
            self.downsample = None

        self.norm1 = self.norm_layer(name="norm1")
        self.act1 = self.act_layer()
        self.conv1 = StdConv2D(
            filters=mid_channels,
            kernel_size=1,
            padding=conv_padding,
            use_bias=False,
            name="conv1",
        )
        self.norm2 = self.norm_layer(name="norm2")
        self.act2 = self.act_layer()
        self.conv2 = StdConv2D(
            filters=mid_channels,
            kernel_size=3,
            strides=strides,
            padding=conv_padding,
            use_bias=False,
            name="conv2",
        )
        self.norm3 = self.norm_layer(name="norm3")
        self.act3 = self.act_layer()
        self.conv3 = StdConv2D(
            filters=nb_channels,
            kernel_size=1,
            padding=conv_padding,
            use_bias=False,
            name="conv3",
        )
        self.drop_path = DropPath(drop_prob=drop_path_rate)

    def call(self, x, training=False):
        # Pre-activation
        y = self.norm1(x, training=training)
        y = self.act1(y)

        if self.downsample is not None:
            shortcut = self.downsample(y, training=training)
        else:
            shortcut = x

        # Residual branch
        x = self.conv1(y)
        x = self.norm2(x, training=training)
        x = self.act2(x)
        x = self.conv2(x)
        x = self.norm3(x, training=training)
        x = self.act3(x)
        x = self.conv3(x)
        x = self.drop_path(x, training=training)

        x = x + shortcut
        return x


class Bottleneck(tf.keras.layers.Layer):
    """
    Non Pre-activation bottleneck block, equiv to V1.5/V1b Bottleneck. Used for ViT.
    """

    def __init__(
        self,
        nb_channels,
        strides: int,
        downsample: bool,
        conv_padding: str,
        act_layer: str,
        norm_layer: str,
        drop_path_rate: float,
        bottleneck_ratio=0.25,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.act_layer = act_layer_factory(act_layer)
        self.norm_layer = norm_layer_factory(norm_layer)
        mid_channels = make_divisible(nb_channels * bottleneck_ratio)

        if downsample:
            self.downsample = DownsampleConv(
                nb_channels=nb_channels,
                strides=strides,
                preact=False,
                conv_padding=conv_padding,
                norm_layer=norm_layer,
                name="downsample",
            )
        else:
            self.downsample = None

        self.conv1 = StdConv2D(
            filters=mid_channels,
            kernel_size=1,
            padding=conv_padding,
            use_bias=False,
            name="conv1",
        )
        self.norm1 = self.norm_layer(name="norm1")
        self.act1 = self.act_layer()
        self.conv2 = StdConv2D(
            filters=mid_channels,
            kernel_size=3,
            strides=strides,
            padding=conv_padding,
            use_bias=False,
            name="conv2",
        )
        self.norm2 = self.norm_layer(name="norm2")
        self.act2 = self.act_layer()
        self.conv3 = StdConv2D(
            filters=nb_channels,
            kernel_size=1,
            padding=conv_padding,
            use_bias=False,
            name="conv3",
        )
        self.norm3 = self.norm_layer(name="norm3")
        self.drop_path = DropPath(drop_prob=drop_path_rate)
        self.act3 = self.act_layer()

    def call(self, x, training=False):
        # Shortcut branch
        if self.downsample is not None:
            shortcut = self.downsample(x, training=training)
        else:
            shortcut = x

        # Residual branch
        x = self.conv1(x)
        x = self.norm1(x, training=training)
        x = self.act1(x)
        x = self.conv2(x)
        x = self.norm2(x, training=training)
        x = self.act2(x)
        x = self.conv3(x)
        x = self.norm3(x, training=training)
        x = self.drop_path(x, training=training)

        x = x + shortcut
        x = self.act3(x)
        return x


class DownsampleConv(tf.keras.layers.Layer):
    def __init__(
        self,
        nb_channels,
        strides,
        preact: bool,
        conv_padding: str,
        norm_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.conv = StdConv2D(
            filters=nb_channels,
            kernel_size=1,
            strides=strides,
            padding=conv_padding,
            use_bias=False,
            name="conv",
        )
        if not preact:
            norm_layer = norm_layer_factory(norm_layer)
            self.norm = norm_layer(name="norm")
        else:
            self.norm = tf.keras.layers.Activation("linear")

    def call(self, x, training=False):
        x = self.conv(x)
        x = self.norm(x, training=training)
        return x


class ResNetV2Stem(tf.keras.layers.Layer):
    def __init__(
        self,
        stem_type: str,
        stem_width: int,
        conv_padding: str,
        preact: bool,
        act_layer: str,
        norm_layer: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.preact = preact

        self.conv = StdConv2D(
            filters=stem_width,
            kernel_size=7,
            strides=2,
            padding=conv_padding,
            use_bias=False,
            name="conv",
        )

        act_layer = act_layer_factory(act_layer)
        norm_layer = norm_layer_factory(norm_layer)
        self.norm = norm_layer(name="norm") if not preact else None
        self.act = act_layer() if not preact else None

        if stem_type == "fixed":
            # 'fixed' SAME padding approximation that is used in BiT models
            self.pad = tf.keras.layers.ZeroPadding2D(padding=1)
            self.pool = tf.keras.layers.MaxPool2D(pool_size=3, strides=2)
        elif stem_type == "same":
            self.pad = tf.keras.layers.Activation("linear")
            self.pool = tf.keras.layers.MaxPool2D(
                pool_size=3, strides=2, padding="same"
            )
        else:
            raise ValueError(f"Unknow value for stem_type: {stem_type}")

    def call(self, x, training=False):
        x = self.conv(x)
        if not self.preact:
            x = self.norm(x, training=training)
            x = self.act(x)
        x = self.pad(x)
        x = self.pool(x)
        return x


@keras_serializable
class ResNetV2(tf.keras.Model):
    """
    Implementation of Pre-activation (v2) ResNet models.
    """

    cfg_class = ResNetV2Config

    def __init__(self, cfg: ResNetV2Config, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.cfg = cfg

        self.stem = ResNetV2Stem(
            stem_type=cfg.stem_type,
            stem_width=make_divisible(cfg.stem_width * cfg.width_factor),
            conv_padding=cfg.conv_padding,
            preact=cfg.preact,
            act_layer=cfg.act_layer,
            norm_layer=cfg.norm_layer,
            name="stem",
        )

        # Stochastic depth
        dpr = np.linspace(0.0, cfg.drop_path_rate, sum(cfg.nb_blocks))
        block_idx = 0

        self.blocks = []
        nb_stages = len(cfg.nb_blocks)
        block_layer = PreActBottleneck if cfg.preact else Bottleneck
        for j in range(nb_stages):
            nb_channels = make_divisible(cfg.nb_channels[j] * cfg.width_factor)
            for k in range(cfg.nb_blocks[j]):
                self.blocks.append(
                    block_layer(
                        nb_channels=nb_channels,
                        strides=2 if (j > 0) & (k == 0) else 1,
                        downsample=k == 0,
                        conv_padding=cfg.conv_padding,
                        act_layer=cfg.act_layer,
                        norm_layer=cfg.norm_layer,
                        drop_path_rate=dpr[block_idx],
                        name=f"stages/{j}/blocks/{k}",
                    )
                )
                block_idx += 1

        if cfg.preact:
            norm_layer = norm_layer_factory(cfg.norm_layer)
            act_layer = act_layer_factory(cfg.act_layer)
            self.norm = norm_layer(name="norm")
            self.act = act_layer()
        else:
            self.norm = None
            self.act = None

        self.head = ClassifierHead(
            nb_classes=cfg.nb_classes,
            pool_type=cfg.global_pool,
            drop_rate=cfg.drop_rate,
            use_conv=False,
            name="head",
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        return (
            ["stem"]
            + [f"block_{j}" for j in range(sum(self.cfg.nb_blocks))]
            + ["features", "logits"]
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x = self.stem(x)
        features["stem"] = x

        for j, block in enumerate(self.blocks):
            x = block(x, training=training)
            features[f"block_{j}"] = x

        if self.cfg.preact:
            x = self.norm(x, training=training)
            x = self.act(x)
        features["features"] = x

        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x, training=training)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def resnetv2_50x1_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_50x1_bitm",
        url="[timm]",
        input_size=(448, 448),
        nb_blocks=(3, 4, 6, 3),
        width_factor=1,
        pool_size=14,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_50x3_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_50x3_bitm",
        url="[timm]",
        input_size=(448, 448),
        nb_blocks=(3, 4, 6, 3),
        width_factor=3,
        pool_size=14,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_101x1_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_101x1_bitm",
        url="[timm]",
        input_size=(448, 448),
        nb_blocks=(3, 4, 23, 3),
        width_factor=1,
        pool_size=14,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_101x3_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_101x3_bitm",
        url="[timm]",
        input_size=(448, 448),
        nb_blocks=(3, 4, 23, 3),
        width_factor=3,
        pool_size=14,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x2_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_152x2_bitm",
        url="[timm]",
        input_size=(448, 448),
        nb_blocks=(3, 8, 36, 3),
        width_factor=2,
        pool_size=14,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x4_bitm():
    cfg = ResNetV2Config(
        name="resnetv2_152x4_bitm",
        url="[timm]",
        input_size=(480, 480),
        nb_blocks=(3, 8, 36, 3),
        width_factor=4,
        pool_size=15,
        crop_pct=1.0,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_50x1_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_50x1_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 4, 6, 3),
        width_factor=1,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_50x3_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_50x3_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 4, 6, 3),
        width_factor=3,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_101x1_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_101x1_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 4, 23, 3),
        width_factor=1,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_101x3_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_101x3_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 4, 23, 3),
        width_factor=3,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x2_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_152x2_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 8, 36, 3),
        width_factor=2,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x4_bitm_in21k():
    cfg = ResNetV2Config(
        name="resnetv2_152x4_bitm_in21k",
        url="[timm]",
        nb_classes=21843,
        nb_blocks=(3, 8, 36, 3),
        width_factor=4,
    )
    return ResNetV2, cfg


@register_model
def resnetv2_50x1_bit_distilled():
    """
    ResNetV2-50x1-BiT Distilled
    Paper: Knowledge distillation: A good teacher is patient and consistent
    Link: https://arxiv.org/abs/2106.05237
    """
    cfg = ResNetV2Config(
        name="resnetv2_50x1_bit_distilled",
        url="[timm]",
        nb_blocks=(3, 4, 6, 3),
        width_factor=1,
        interpolation="bicubic",
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x2_bit_teacher():
    """
    ResNetV2-152x2-BiT Teacher
    Paper: Knowledge distillation: A good teacher is patient and consistent
    Link: https://arxiv.org/abs/2106.05237
    """
    cfg = ResNetV2Config(
        name="resnetv2_152x2_bit_teacher",
        url="[timm]",
        nb_blocks=(3, 8, 36, 3),
        width_factor=2,
        interpolation="bicubic",
    )
    return ResNetV2, cfg


@register_model
def resnetv2_152x2_bit_teacher_384():
    """
    ResNetV2-152x2-BiT Teacher @ 384x384
    Paper: Knowledge distillation: A good teacher is patient and consistent
    Link: https://arxiv.org/abs/2106.05237
    """
    cfg = ResNetV2Config(
        name="resnetv2_152x2_bit_teacher_384",
        url="[timm]",
        input_size=(384, 384),
        nb_blocks=(3, 8, 36, 3),
        width_factor=2,
        pool_size=12,
        crop_pct=1.0,
        interpolation="bicubic",
    )
    return ResNetV2, cfg



"""
Pooling and classifier head.

Ported to TF from timm/models/layers/classifier.py by Ross Wightman.

Copyright 2021 Martins Bruveris
"""
import tensorflow as tf


class ClassifierHead(tf.keras.layers.Layer):
    """Classifier head with configurable global pooling and dropout."""

    def __init__(
        self,
        nb_classes: int,
        pool_type: str = "avg",
        drop_rate: float = 0.0,
        use_conv: bool = False,
        name: str = "head",
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)

        pool_type = pool_type or ""  # Convert other falsy values to empty string
        if not pool_type and not use_conv and nb_classes > 0:
            raise ValueError(
                "Pooling can only be disabled if conv classifier is used "
                "or classifier is disabled."
            )
        if pool_type == "":
            self.pool = None
        elif pool_type == "avg":
            self.pool = tf.keras.layers.GlobalAveragePooling2D()
        elif pool_type == "max":
            self.pool = tf.keras.layers.GlobalMaxPool2D()
        else:
            raise NotImplementedError(f"pool_type={pool_type} not implemented.")

        if drop_rate > 0.0:
            self.drop = tf.keras.layers.Dropout(rate=drop_rate)
        else:
            self.drop = None

        if nb_classes == 0:
            self.fc = None
        elif use_conv:
            self.fc = tf.keras.layers.Conv2D(
                filters=nb_classes,
                kernel_size=1,
                use_bias=True,
                name="fc",
            )
        else:
            self.fc = tf.keras.layers.Dense(
                units=nb_classes,
                use_bias=True,
                name="fc",
            )
        if pool_type == "":
            self.flatten = None
        else:
            self.flatten = tf.keras.layers.Flatten()

    def call(self, x, training=False):
        if self.pool is not None:
            x = self.pool(x)
        if self.drop is not None:
            x = self.drop(x, training=training)
        if self.fc is not None:
            x = self.fc(x)
        if self.flatten is not None:
            x = self.flatten(x)
        return x



"""
TensorFlow implementation of ConvMixer

Based on timm/models/convmixer.py by Ross Wightman.

Copyright 2021 Martins Bruveris
"""
from dataclasses import dataclass
from typing import List, Tuple

import tensorflow as tf

from tfimm.layers import act_layer_factory, norm_layer_factory
from tfimm.models import ModelConfig, keras_serializable, register_model
from tfimm.utils import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD

# model_registry will add each entrypoint fn to this
__all__ = ["ConvMixer", "ConvMixerConfig"]


@dataclass
class ConvMixerConfig(ModelConfig):
    nb_classes: int = 1000
    in_channels: int = 3
    input_size: Tuple[int, int] = (224, 224)
    patch_size: Tuple[int, int] = (7, 7)
    embed_dim: int = 768
    depth: int = 32
    kernel_size: int = 9
    norm_layer: str = "batch_norm"
    act_layer: str = "gelu"
    # Parameters for inference
    crop_pct: float = 0.96
    interpolation: str = "bicubic"
    mean: Tuple[float, float, float] = IMAGENET_DEFAULT_MEAN
    std: Tuple[float, float, float] = IMAGENET_DEFAULT_STD
    first_conv: str = "stem/0"
    classifier: str = "head"


class Block(tf.keras.layers.Layer):
    def __init__(self, cfg: ConvMixerConfig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        act_layer = act_layer_factory(cfg.act_layer)
        norm_layer = norm_layer_factory(cfg.norm_layer)

        self.conv1 = tf.keras.layers.DepthwiseConv2D(
            kernel_size=cfg.kernel_size,
            padding="same",
            name="0/fn/0",
        )
        self.act1 = act_layer()
        self.bn1 = norm_layer(name="0/fn/2")

        self.conv2 = tf.keras.layers.Conv2D(
            filters=cfg.embed_dim,
            kernel_size=1,
            name="1",
        )
        self.act2 = act_layer()
        self.bn2 = norm_layer(name="3")

    def call(self, x, training=False):
        x_residual = x
        x = self.conv1(x)
        x = self.act1(x)
        x = self.bn1(x, training=training)
        x = x + x_residual

        x = self.conv2(x)
        x = self.act2(x)
        x = self.bn2(x, training=training)
        return x


@keras_serializable
class ConvMixer(tf.keras.Model):
    cfg_class = ConvMixerConfig

    def __init__(self, cfg: ConvMixerConfig, *args, **kwargs):
        kwargs["name"] = kwargs.get("name", cfg.name)
        super().__init__(*args, **kwargs)
        self.nb_features = cfg.embed_dim  # For consistency with other models
        self.act_layer = act_layer_factory(cfg.act_layer)
        self.norm_layer = norm_layer_factory(cfg.norm_layer)
        self.cfg = cfg

        conv1 = tf.keras.layers.Conv2D(
            filters=cfg.embed_dim,
            kernel_size=cfg.patch_size,
            strides=cfg.patch_size,
            name=f"{self.name}/stem/0",
        )
        act1 = self.act_layer()
        bn1 = self.norm_layer(name=f"{self.name}/stem/2")
        self.stem = tf.keras.Sequential([conv1, act1, bn1])

        self.blocks = [Block(cfg, name=f"blocks/{j}") for j in range(cfg.depth)]
        self.pool = tf.keras.layers.GlobalAveragePooling2D()
        self.flatten = tf.keras.layers.Flatten()

        self.head = (
            tf.keras.layers.Dense(units=cfg.nb_classes, name="head")
            if cfg.nb_classes > 0
            else tf.keras.layers.Activation("linear")
        )

    @property
    def dummy_inputs(self) -> tf.Tensor:
        return tf.zeros((1, *self.cfg.input_size, self.cfg.in_channels))

    @property
    def feature_names(self) -> List[str]:
        return (
            ["stem"]
            + [f"block_{j}" for j in range(self.cfg.depth)]
            + ["features_all", "features", "logits"]
        )

    def forward_features(self, x, training=False, return_features=False):
        features = {}
        x = self.stem(x, training=training)
        features["stem"] = x
        for j, block in enumerate(self.blocks):
            # noinspection PyCallingNonCallable
            x = block(x, training=training)
            features[f"block_{j}"] = x
        features["features_all"] = x
        x = self.pool(x)
        x = self.flatten(x)
        features["features"] = x
        return (x, features) if return_features else x

    def call(self, x, training=False, return_features=False):
        features = {}
        x = self.forward_features(x, training, return_features)
        if return_features:
            x, features = x
        x = self.head(x)
        features["logits"] = x
        return (x, features) if return_features else x


@register_model
def convmixer_768_32():
    """
    ConvMixer-768/32.
    Source: https://github.com/tmp-iclr/convmixer
    Note: This network uses ReLU activation instead of GELU
    """
    cfg = ConvMixerConfig(
        name="convmixer_768_32",
        url="[timm]",
        patch_size=(7, 7),
        embed_dim=768,
        depth=32,
        kernel_size=7,
        act_layer="relu",
    )
    return ConvMixer, cfg


@register_model
def convmixer_1024_20_ks9_p14():
    """
    ConvMixer-1024/20.
    Source: https://github.com/tmp-iclr/convmixer
    """
    cfg = ConvMixerConfig(
        name="convmixer_1024_20_ks9_p14",
        url="[timm]",
        patch_size=(14, 14),
        embed_dim=1024,
        depth=20,
        kernel_size=9,
    )
    return ConvMixer, cfg


@register_model
def convmixer_1536_20():
    """
    ConvMixer-1536/20.
    Source: https://github.com/tmp-iclr/convmixer
    """
    cfg = ConvMixerConfig(
        name="convmixer_1536_20",
        url="[timm]",
        patch_size=(7, 7),
        embed_dim=1536,
        depth=20,
        kernel_size=9,
    )
    return ConvMixer, cfg



import tensorflow as tf

from tfimm.layers import act_layer_factory


class MLPBlock(tf.keras.layers.Layer):
    """MLP block. Only difference with tfimm version is naming of layers."""

    def __init__(
        self,
        hidden_dim: int,
        embed_dim: int,
        drop_rate: float,
        act_layer: str,
        kernel_initializer: str = "glorot_uniform",
        bias_initializer: str = "zeros",
        **kwargs,
    ):
        super().__init__(**kwargs)
        act_layer = act_layer_factory(act_layer)

        self.fc1 = tf.keras.layers.Dense(
            units=hidden_dim,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="lin1",
        )
        self.act = act_layer()
        self.drop1 = tf.keras.layers.Dropout(rate=drop_rate)
        self.fc2 = tf.keras.layers.Dense(
            units=embed_dim,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            name="lin2",
        )
        self.drop2 = tf.keras.layers.Dropout(rate=drop_rate)

    def call(self, x, training=False):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x, training=training)
        x = self.fc2(x)
        x = self.drop2(x, training=training)
        return x



from dataclasses import dataclass
from typing import Any

import tensorflow as tf
from tensorflow.keras.optimizers.schedules import LearningRateSchedule

from ..registry import cfg_serializable, get_class


@dataclass
class OptimizerConfig:
    # Lr schedule. Currently, supports `const`, `multisteps`, `cosine_decay` and
    # `exponential_decay`. Different schedules will use different sets of parameters
    # from below.
    lr_schedule: Any
    lr_schedule_class: str

    # Linear lr warmup over `lr_warmup` number of epochs from 0 to the value specified
    # by the schedule.
    lr_warmup: int = -1

    # Which optimizer to use. Currently supports `sgd`, `adam`, `rmsprop`, `adadelta`,
    # `adamax` and `adagrad`.
    optimizer: str = "sgd"
    # Parameters for the optimizer. Usually momentum values. Not all optimizers need
    # both betas, e.g., `sgd` only uses `betas[0]` for its momentum. But, for
    # consistency, we always need to pass a tuple for `betas`.
    betas: tuple = (0.9, 0.999)
    # Gradient clipping parameters. `clipnorm` clips gradients using the l2-norm (on
    # each weight), while `clipvalue` clips each entry independently. Internally TF
    # applies `tf.clip_by_norm` or `tf.clip_by_value` on each weight.
    clipnorm: float = -1.0
    clipvalue: float = -1.0
    # Epsilon parameter for Adam and RMSProp
    epsilon: float = 1e-7
    # Rho parameter representing the decay rate for Adadelta
    rho: float = 0.95
    # Initial accumulator value for Adagrad. Starting value for the accumulators
    # (per-parameter momentum values). Must be non-negative.
    initial_accumulator_value: float = 0.1


@cfg_serializable
class OptimizerFactory:
    cfg_class = OptimizerConfig

    def __init__(self, cfg: OptimizerConfig, timekeeping, mixed_precision: bool):
        self.cfg = cfg
        self.timekeeping = timekeeping
        self.mixed_precision = mixed_precision

    def lr_schedule(self):
        """Create learning rate schedule as defined by config."""

        lr = get_class(self.cfg.lr_schedule_class)(
            cfg=self.cfg.lr_schedule,
            timekeeping=self.timekeeping,
        )()

        if self.cfg.lr_warmup != -1:
            lr = WarmupWrapper(
                lr_schedule=lr,
                warmup_steps=self.cfg.lr_warmup * self.timekeeping.nb_steps_per_epoch,
            )

        return lr

    def optimizer(self, lr):
        cfg = self.cfg

        if cfg.clipnorm != -1.0 and cfg.clipvalue != -1.0:
            raise ValueError(
                "`clipnorm` and `clipvalue` cannot both be used simultaneously."
            )

        # We cannot use `None` in the config class, because we want to adhere to typing
        # to make parsing of configs easier. But TF expects `None`.
        clipnorm = cfg.clipnorm if cfg.clipnorm != -1.0 else None
        clipvalue = cfg.clipvalue if cfg.clipvalue != -1.0 else None

        if cfg.optimizer == "sgd":
            opt = tf.keras.optimizers.SGD(
                learning_rate=lr,
                momentum=cfg.betas[0],
                clipnorm=clipnorm,
                clipvalue=clipvalue,
            )
        elif cfg.optimizer == "adam":
            opt = tf.keras.optimizers.Adam(
                learning_rate=lr,
                beta_1=cfg.betas[0],
                beta_2=cfg.betas[1],
                clipnorm=clipnorm,
                clipvalue=clipvalue,
                epsilon=cfg.epsilon,
            )
        elif cfg.optimizer == "rmsprop":
            opt = tf.keras.optimizers.RMSprop(
                learning_rate=lr,
                rho=cfg.betas[0],
                momentum=cfg.betas[1],
                clipnorm=clipnorm,
                clipvalue=clipvalue,
                epsilon=cfg.epsilon,
            )
        elif cfg.optimizer == "adamax":
            opt = tf.keras.optimizers.Adamax(
                learning_rate=lr,
                beta_1=cfg.betas[0],
                beta_2=cfg.betas[1],
                clipnorm=clipnorm,
                clipvalue=clipvalue,
                epsilon=cfg.epsilon,
            )
        elif cfg.optimizer == "adadelta":
            opt = tf.keras.optimizers.Adadelta(
                learning_rate=lr,
                rho=cfg.rho,
                clipnorm=clipnorm,
                clipvalue=clipvalue,
                epsilon=cfg.epsilon,
            )
        elif cfg.optimizer == "adagrad":
            opt = tf.keras.optimizers.Adagrad(
                learning_rate=lr,
                initial_accumulator_value=cfg.initial_accumulator_value,
                clipnorm=clipnorm,
                clipvalue=clipvalue,
                epsilon=cfg.epsilon,
            )
        else:
            raise ValueError(f"Unknown optimizer: {cfg.optimizer}.")

        if self.mixed_precision:
            opt = tf.keras.mixed_precision.LossScaleOptimizer(opt, dynamic=True)

        return opt

    def __call__(self):
        lr = self.lr_schedule()
        opt = self.optimizer(lr)
        return opt


class WarmupWrapper(LearningRateSchedule):
    """Implements linear learning rate warmup."""

    def __init__(self, lr_schedule: LearningRateSchedule, warmup_steps: int):
        self.lr_schedule = lr_schedule
        self.warmup_steps = warmup_steps
        # Warmup will linearly increase the lr from 0 to the value we would have if
        # we followed the lr schedule.
        self.target_lr = lr_schedule(warmup_steps)

    def __call__(self, step):
        warmup_lr = self.target_lr * tf.cast(step / self.warmup_steps, tf.float32)
        # We can't use an if/else statement, because the code is compiled in graph mode.
        lr = tf.cond(
            step < self.warmup_steps,
            true_fn=lambda: warmup_lr,
            false_fn=lambda: self.lr_schedule(step),
        )
        return lr

    def get_config(self):
        return {
            "warmup_steps": self.warmup_steps,
            "lr_schedule": tf.keras.optimizers.schedules.serialize(self.lr_schedule),
        }

    @classmethod
    def from_config(cls, config):
        return WarmupWrapper(
            lr_schedule=tf.keras.optimizers.schedules.deserialize(
                config["lr_schedule"]
            ),
            warmup_steps=config["warmup_steps"],
        )
